#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from enum import IntEnum
from glob import glob
import json
import os
import re
from typing import Dict
from typing import List
from typing import Optional

import networkx as nx
from traitlets.config import SingletonConfigurable

from elyra.pipeline.component import Component
from elyra.pipeline.component_registry import ComponentRegistry
from elyra.pipeline.pipeline import Operation
from elyra.pipeline.pipeline import PIPELINE_CURRENT_SCHEMA
from elyra.pipeline.pipeline import PIPELINE_CURRENT_VERSION
from elyra.pipeline.pipeline_definition import PipelineDefinition
from elyra.pipeline.processor import PipelineProcessorManager
from elyra.util.path import get_expanded_path


class ValidationSeverity(IntEnum):
    Error = 1
    Warning = 2
    Information = 3
    Hint = 4


class ValidationResponse(object):
    def __init__(self):
        self._response = {"title": "Elyra Pipeline Diagnostics",
                          "description": "Issues discovered when parsing the pipeline",
                          "issues": []
                          }
        self._has_fatal = False

    @property
    def response(self) -> Dict:
        """
        :return: The dict of validation errors and warnings found in the pipeline
        """
        return self._response

    @property
    def has_fatal(self):
        return self._has_fatal

    def add_message(self,
                    message: str,
                    message_type: Optional[str] = "",
                    data: Optional[Dict] = "",
                    severity: ValidationSeverity = ValidationSeverity.Warning):
        """
        Helper function to add a diagnostic message to the response to be sent back
        :param message: A simple message describing the issue
        :param message_type: The type of message to send back e.g. invalidNodeType, invalidPipeline
        :param data: a Dict with granular details regarding the error e.g. the nodeID, pipelineID, linkID etc.
        :param severity: the severity level of the issue
        :return:
        """
        valid_severity_levels = [ValidationSeverity.Error, ValidationSeverity.Warning,
                                 ValidationSeverity.Information, ValidationSeverity.Hint]

        if severity in valid_severity_levels:
            diagnostic = {"severity": severity.value,
                          "source": "Elyra Pipeline Validation Service",
                          "type": message_type,
                          "message": message,
                          "data": data
                          }
            self._response['issues'].append(diagnostic)

        if severity is ValidationSeverity.Error:
            self._has_fatal = True

    def to_json(self):
        return self._response


class PipelineValidationManager(SingletonConfigurable):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.root_dir = get_expanded_path(kwargs.get('root_dir'))

    async def validate(self, pipeline: Dict) -> ValidationResponse:
        """
        Validates the pipeline JSON payload
        :param pipeline: the pipeline definition to be validated
        :return: ValidationResponse containing any and all issues discovered during the validation
        """
        response = ValidationResponse()

        pipeline_definition = PipelineDefinition(pipeline_definition=pipeline)
        issues = pipeline_definition.validate()
        for issue in issues:
            response.add_message(severity=ValidationSeverity.Error,
                                 message_type="invalidJSON",
                                 message=issue)

        try:
            pipeline_definition.primary_pipeline
        except ValueError:
            response.add_message(severity=ValidationSeverity.Error,
                                 message_type="invalidJSON",
                                 message="Invalid JSON detected, unable to continue.")

            return response

        pipeline_runtime = pipeline_definition.primary_pipeline.runtime  # local, kfp, airflow
        if PipelineProcessorManager.instance().is_supported_runtime(pipeline_runtime) is False:
            response.add_message(severity=ValidationSeverity.Error,
                                 message_type="invalidRuntime",
                                 message="Unsupported pipeline runtime",
                                 data={"pipelineRuntime": pipeline_runtime})

        pipeline_type = pipeline_definition.primary_pipeline.type  # generic, kfp, airflow
        if pipeline_type == 'generic' and \
                PipelineProcessorManager.instance().is_supported_runtime(pipeline_runtime) is False:
            response.add_message(severity=ValidationSeverity.Error,
                                 message_type="invalidRuntime",
                                 message="Unsupported pipeline type",
                                 data={"pipelineRuntime": pipeline_type})

        self._validate_pipeline_structure(pipeline_definition=pipeline_definition,
                                          response=response)

        await self._validate_compatibility(pipeline_definition=pipeline_definition,
                                           pipeline_type=pipeline_type,
                                           pipeline_runtime=pipeline_runtime,
                                           response=response)

        self._validate_pipeline_graph(pipeline=pipeline, response=response)

        if response.has_fatal:
            return response

        await self._validate_node_properties(pipeline_definition=pipeline_definition,
                                             pipeline_type=pipeline_type,
                                             pipeline_runtime=pipeline_runtime,
                                             response=response)

        return response

    def _validate_pipeline_structure(self, pipeline_definition: PipelineDefinition,
                                     response: ValidationResponse) -> None:
        """
        Validates the pipeline structure based on version of schema
        :param pipeline_definition: the pipeline definition to be validated
        :param response: ValidationResponse containing the issue list to be updated
        """

        # Validate pipeline schema version
        if float(pipeline_definition.schema_version) != PIPELINE_CURRENT_SCHEMA:
            response.add_message(severity=ValidationSeverity.Error, message_type="invalidPipeline",
                                 message="Incompatible pipeline schema version detected.",
                                 data={"supported_schema_version": PIPELINE_CURRENT_SCHEMA,
                                       "detected_schema_version": float(pipeline_definition.schema_version)})

        # validate pipeline version compatibility
        pipeline_version = int(pipeline_definition.primary_pipeline.version)
        if pipeline_version not in range(PIPELINE_CURRENT_VERSION + 1):
            response.add_message(severity=ValidationSeverity.Error, message_type="invalidPipeline",
                                 message="Primary pipeline version field has an invalid value.",
                                 data={"supported_version": PIPELINE_CURRENT_VERSION,
                                       "detected_version": pipeline_version})

        elif pipeline_version < PIPELINE_CURRENT_VERSION:
            # Pipeline needs to be migrated
            response.add_message(severity=ValidationSeverity.Error, message_type="invalidPipeline",
                                 message=f'Pipeline version {pipeline_version} is out of date and needs to be migrated '
                                         f'using the Elyra pipeline editor.')
        elif pipeline_version > PIPELINE_CURRENT_VERSION:
            # New version of Elyra is needed
            response.add_message(severity=ValidationSeverity.Error, message_type="invalidPipeline",
                                 message='Pipeline was last edited in a newer version of Elyra. '
                                         'Update Elyra to use this pipeline.',
                                 data={"supported_version": PIPELINE_CURRENT_VERSION,
                                       "detected_version": pipeline_version})

    async def _validate_compatibility(self, pipeline_definition: PipelineDefinition,
                                      pipeline_type: str,
                                      pipeline_runtime: str,
                                      response: ValidationResponse) -> None:
        """
        Checks that the pipeline payload is compatible with this version of elyra (ISSUE #938)
        as well as verifying all nodes in the pipeline are supported by the runtime
        :param pipeline_definition: the pipeline definition to be validated
        :param pipeline_type: name of the pipeline runtime being used e.g. kfp, airflow, generic
        :param pipeline_runtime: name of the pipeline runtime for execution  e.g. kfp, airflow, local
        :param response: ValidationResponse containing the issue list to be updated
        """

        primary_pipeline_id = pipeline_definition.primary_pipeline.id
        supported_ops = []

        if pipeline_runtime:
            if pipeline_runtime != pipeline_type and pipeline_type != 'generic':
                response.add_message(severity=ValidationSeverity.Error,
                                     message_type="invalidRuntime",
                                     message="Pipeline runtime platform is not compatible "
                                             "with selected runtime configuration.",
                                     data={"pipelineID": primary_pipeline_id,
                                           "pipelineType": pipeline_type,
                                           "pipelineRuntime": pipeline_runtime})
            elif PipelineProcessorManager.instance().is_supported_runtime(pipeline_runtime):
                component_list = await PipelineProcessorManager.instance().get_components(pipeline_runtime)
                for component in component_list:
                    supported_ops.append(component.op)

                # Checks pipeline node types are compatible with the runtime selected
                for sub_pipeline in pipeline_definition.pipelines:
                    for node in sub_pipeline.nodes:
                        if node.type == "execution_node" and node.op not in supported_ops:
                            response.add_message(severity=ValidationSeverity.Error,
                                                 message_type="invalidNodeType",
                                                 message="This component was not found in the registry. Please add it "
                                                         "to your component registry or remove this node from the "
                                                         "pipeline",
                                                 data={"nodeID": node.id,
                                                       "nodeOpName": node.op,
                                                       "nodeName": node.label,
                                                       "pipelineId": sub_pipeline.id})
            else:
                response.add_message(severity=ValidationSeverity.Error,
                                     message_type="invalidRuntime",
                                     message="Unsupported pipeline runtime",
                                     data={"pipelineRuntime": pipeline_runtime,
                                           "pipelineType": pipeline_type,
                                           "pipelineId": primary_pipeline_id})

    async def _validate_node_properties(self, pipeline_definition: PipelineDefinition,
                                        pipeline_type: str,
                                        pipeline_runtime: str,
                                        response: ValidationResponse) -> None:
        """
        Validates each of the node's structure for required fields/properties as well as
        their values
        :param pipeline_definition: the pipeline definition to be validated
        :param pipeline_type: name of the pipeline runtime being used e.g. kfp, airflow, generic
        :param pipeline_runtime: name of the pipeline runtime for execution  e.g. kfp, airflow, local
        :param response: ValidationResponse containing the issue list to be updated
        """
        if pipeline_runtime:
            # don't check if incompatible pipeline type and runtime
            if pipeline_runtime != pipeline_type and pipeline_type != 'generic':
                return

        for pipeline in pipeline_definition.pipelines:
            component_list = await PipelineProcessorManager.instance().get_components(pipeline_runtime)
            components = ComponentRegistry.to_canvas_palette(component_list)
            for node in pipeline.nodes:
                if node.type == 'execution_node':
                    node_label = node.label
                    if Operation.is_generic_operation(node.op):
                        image_name = node.get_component_parameter('runtime_image')
                        filename = node.get_component_parameter("filename")
                        dependencies = node.get_component_parameter("dependencies")
                        env_vars = node.get_component_parameter("env_vars")

                        self._validate_filepath(node_id=node.id, node_label=node_label, property_name='filename',
                                                filename=filename, response=response)

                        # If not running locally, we check resource and image name
                        if pipeline_runtime != 'local':
                            self._validate_container_image_name(node.id, node_label, image_name, response=response)
                            for resource_name in ['cpu', 'gpu', 'memory']:
                                resource_value = node.get_component_parameter(resource_name)
                                if resource_value:
                                    self._validate_resource_value(node.id, node_label, resource_name=resource_name,
                                                                  resource_value=resource_value,
                                                                  response=response)

                        self._validate_label(node_id=node.id, node_label=node_label, response=response)
                        if dependencies:
                            notebook_root_relative_path = os.path.dirname(filename)
                            for dependency in dependencies:
                                self._validate_filepath(node_id=node.id, node_label=node_label,
                                                        file_dir=os.path.join(self.root_dir,
                                                                              notebook_root_relative_path),
                                                        property_name='dependencies',
                                                        filename=dependency, response=response)
                        if env_vars:
                            for env_var in env_vars:
                                self._validate_environmental_variables(node.id, node_label, env_var=env_var,
                                                                       response=response)

                    # Validate runtime components against specific node properties in component registry
                    else:
                        # This is the full dict of properties for the operation e.g. current params, optionals etc
                        property_dict = await self._get_component_properties(pipeline_type, components, node.op)
                        cleaned_property_list = list(map(lambda x: str(x).replace('elyra_', ''),
                                                         property_dict['current_parameters'].keys()))

                        # Remove the non component_parameter jinja templated values we do not check against
                        cleaned_property_list.remove('component_source')
                        cleaned_property_list.remove('label')

                        for node_property in cleaned_property_list:
                            if not node.get_component_parameter(node_property):
                                if self._is_required_property(property_dict, node_property):
                                    response.add_message(severity=ValidationSeverity.Error,
                                                         message_type="invalidNodeProperty",
                                                         message="Node is missing required property.",
                                                         data={"nodeID": node.id,
                                                               "nodeName": node_label,
                                                               "propertyName": node_property})
                            elif self._get_component_type(property_dict, node_property) == 'inputpath':
                                if len(node.get_component_parameter(node_property).keys()) < 2:
                                    response.add_message(severity=ValidationSeverity.Error,
                                                         message_type="invalidNodeProperty",
                                                         message="Node is missing required output property parameter",
                                                         data={"nodeID": node.id,
                                                               "nodeName": node_label})
                                else:
                                    for key in node.get_component_parameter(node_property).keys():
                                        if key not in ['value', 'option']:
                                            response.add_message(severity=ValidationSeverity.Error,
                                                                 message_type="invalidNodeProperty",
                                                                 message="Node property has invalid key.",
                                                                 data={"nodeID": node.id,
                                                                       "nodeName": node_label,
                                                                       "propertyName": node_property,
                                                                       "keyName": key})

                                node_ids = list(x.get('node_id_ref', None) for x in node.component_links)
                                parent_list = self._get_parent_id_list(pipeline_definition, node_ids, [])
                                if node.get_component_parameter(node_property)['value'] not in parent_list:
                                    response.add_message(severity=ValidationSeverity.Error,
                                                         message_type="invalidNodeProperty",
                                                         message="Node contains an invalid inputpath reference. Please "
                                                                 "check your node-to-node connections",
                                                         data={"nodeID": node.id,
                                                               "nodeName": node_label})

    def _validate_container_image_name(self, node_id: str, node_label: str, image_name: str,
                                       response: ValidationResponse) -> None:
        """
        Validates the image name exists and is proper in syntax
        :param node_id: the unique ID of the node
        :param node_label: the given node name or user customized name/label of the node
        :param image_name: container image name to be evaluated
        :param response: ValidationResponse containing the issue list to be updated
        """
        if not image_name:
            response.add_message(severity=ValidationSeverity.Error,
                                 message_type="invalidNodeProperty",
                                 message="Required property value is missing.",
                                 data={"nodeID": node_id,
                                       "nodeName": node_label,
                                       "propertyName": 'runtime_image'})

    def _validate_resource_value(self, node_id: str, node_label: str, resource_name: str,
                                 resource_value: str, response: ValidationResponse) -> None:
        """
        Validates the value for hardware resources requested
        :param node_id: the unique ID of the node
        :param node_label: the given node name or user customized name/label of the node
        :param resource_name: the name of the resource e.g. cpu, gpu. memory
        :param resource_value: the value of the resource
        :param response: ValidationResponse containing the issue list to be updated
        """
        try:
            if int(resource_value) <= 0:
                response.add_message(severity=ValidationSeverity.Error,
                                     message_type="invalidNodeProperty",
                                     message="Property must be greater than zero.",
                                     data={"nodeID": node_id,
                                           "nodeName": node_label,
                                           "propertyName": resource_name,
                                           "value": resource_value})
        except (ValueError, TypeError):
            response.add_message(severity=ValidationSeverity.Error,
                                 message_type="invalidNodeProperty",
                                 message="Property has a non-numeric value.",
                                 data={"nodeID": node_id,
                                       "nodeName": node_label,
                                       "propertyName": resource_name,
                                       "value": resource_value})

    def _validate_filepath(self, node_id: str, node_label: str, property_name: str,
                           filename: str, response: ValidationResponse, file_dir: Optional[str] = "") -> None:
        """
        Checks the file structure, paths and existence of pipeline dependencies.
        Note that this does not cross reference with file path references within the notebook or script itself.
        :param node_id: the unique ID of the node
        :param node_label: the given node name or user customized name/label of the node
        :param property_name: name of the node property being validated
        :param filename: the name of the file or directory to verify
        :param response: ValidationResponse containing the issue list to be updated
        :param file_dir: the dir path of the where the pipeline file resides in the elyra workspace
        """
        file_dir = file_dir or self.root_dir

        if filename == os.path.abspath(filename):
            normalized_path = os.path.normpath(filename)
        elif filename.startswith(file_dir):
            normalized_path = os.path.normpath(filename)
        else:
            normalized_path = os.path.normpath(f"{os.path.join(file_dir, filename)}")

        if not os.path.commonpath([normalized_path, self.root_dir]) == self.root_dir:
            response.add_message(severity=ValidationSeverity.Error,
                                 message_type="invalidFilePath",
                                 message="Property has an invalid reference to a file/dir outside the root workspace.",
                                 data={"nodeID": node_id,
                                       "nodeName": node_label,
                                       "propertyName": property_name,
                                       "value": normalized_path})
        elif '*' in normalized_path:
            if len(glob(normalized_path)) == 0:
                response.add_message(severity=ValidationSeverity.Error,
                                     message_type="invalidFilePath",
                                     message="Property(wildcard) has an invalid path to a file/dir"
                                             " or the file/dir does not exist.",
                                     data={"nodeID": node_id,
                                           "nodeName": node_label,
                                           "propertyName": property_name,
                                           "value": normalized_path})
        elif not os.path.exists(normalized_path):
            response.add_message(severity=ValidationSeverity.Error,
                                 message_type="invalidFilePath",
                                 message="Property has an invalid path to a file/dir or the file/dir does not exist.",
                                 data={"nodeID": node_id,
                                       "nodeName": node_label,
                                       "propertyName": property_name,
                                       "value": normalized_path})

    def _validate_environmental_variables(self, node_id: str, node_label: str,
                                          env_var: str, response: ValidationResponse) -> None:
        """
        Checks the format of the env var to ensure its in the correct form
        e.g. FOO = 'BAR'
        :param node_id: the unique ID of the node
        :param node_label: the given node name or user customized name/label of the node
        :param env_var: the env_var key value pair to check
        :param response: ValidationResponse containing the issue list to be updated
        """
        result = [x.strip(' \'\"') for x in env_var.split('=', 1)]
        if len(result) != 2:
            response.add_message(severity=ValidationSeverity.Error,
                                 message_type="invalidEnvPair",
                                 message="Property has an improperly formatted env variable key value pair.",
                                 data={"nodeID": node_id,
                                       "nodeName": node_label,
                                       "propertyName": 'env_vars',
                                       "value": env_var})

    def _validate_label(self, node_id: str, node_label: str, response: ValidationResponse) -> None:
        """
        KFP specific check for the label name when constructing the node operation using dsl
        :param node_id: the unique ID of the node
        :param node_label: the given node name or user customized name/label of the node
        :param response: ValidationResponse containing the issue list to be updated
        """
        label_name_max_length = 63
        label_regex = re.compile('^[a-z0-9]([-_.a-z0-9]{0,62}[a-z0-9])?')
        matched = label_regex.search(node_label)

        if len(node_label) > label_name_max_length:
            response.add_message(severity=ValidationSeverity.Warning,
                                 message_type="invalidNodeLabel",
                                 message="Property value exceeds the max length allowed "
                                         "({label_name_max_length}). This value may be truncated "
                                         "by the runtime service.",
                                 data={"nodeID": node_id,
                                       "nodeName": node_label,
                                       "propertyName": 'label',
                                       "value": node_label})
        if not matched or matched.group(0) != node_label:
            response.add_message(severity=ValidationSeverity.Warning,
                                 message_type="invalidNodeLabel",
                                 message="The node label contains characters that may be replaced "
                                         "by the runtime service. Node labels should "
                                         "start with lower case alphanumeric and contain "
                                         "only lower case alphanumeric, underscores, dots, and dashes.",
                                 data={"nodeID": node_id,
                                       "nodeName": node_label,
                                       "propertyName": 'label',
                                       "value": node_label})

    def _validate_pipeline_graph(self, pipeline: dict, response: ValidationResponse) -> None:
        """
        Validates that the pipeline is an acyclic graph, meaning no circular references
        Converts the pipeline definition into a series of tuple node edges(arrows) that represent the connections
        from one node to another via a networkX DiGraph.
        Example:
                 NodeC
                   ^
                   |
        NodeA -> NodeB -> NodeD
                   ^         |
                   |         |    (Invalid Circular Reference)
                   <---------

        The resulting list of edges (arrows) would then be:
        [(NodeA, NodeB), (NodeB, NodeC), (NodeB, NodeD), (NodeD, NodeB)]

        the list of nodes added would be:
        [NodeA, NodeB, NodeC, NodeD]

        This function will add an error message for each cycle found and provide a list of LinkID(s)
        representing the cycle, in the example above, we would return a single error message with the LinkIDs
        for (NodeB, NodeD) and (NodeD, NodeB)
        :param response: ValidationResponse containing the issue list to be updated
        :param pipeline: A dictionary describing the pipeline
        """
        pipeline_json = json.loads(json.dumps(pipeline))

        graph = nx.DiGraph()

        for single_pipeline in pipeline_json['pipelines']:
            node_list = single_pipeline['nodes']
            for node in node_list:
                if node['type'] == "execution_node":
                    graph.add_node(node['id'])
                    if node.get('inputs'):
                        if "links" in node['inputs'][0]:
                            for link in node['inputs'][0]['links']:
                                if "_outPort" in link['port_id_ref']:  # is ref to node, doesnt add links to supernodes
                                    graph.add_edge(link['port_id_ref'].strip('_outPort'), node['id'])
                                elif link['port_id_ref'] == "outPort":  # do not link to bindings
                                    graph.add_edge(link['node_id_ref'], node['id'])
                if node['type'] == "super_node":
                    for link in node['inputs'][0]['links']:
                        child_node_id = node['inputs'][0]['id'].strip("_inPort")
                        graph.add_edge(link['node_id_ref'], child_node_id)

        for isolate in nx.isolates(graph):
            if graph.number_of_nodes() > 1:
                response.add_message(severity=ValidationSeverity.Warning,
                                     message_type="singletonReference",
                                     message="Node is not connected to any other node.",
                                     data={"nodeID": isolate,
                                           "nodeName":
                                           self._get_node_names(pipeline=pipeline, node_id_list=[isolate])[0],
                                           "pipelineID": self._get_pipeline_id(pipeline, node_id=isolate)})

        cycles_detected = nx.simple_cycles(graph)

        if len(list(cycles_detected)) > 0:
            response.add_message(severity=ValidationSeverity.Error,
                                 message_type="circularReference",
                                 message="The pipeline contains a circular dependency between nodes.",
                                 data={})

    def _get_pipeline_id(self, pipeline: dict, node_id: str) -> str:
        """
        Given a node ID, returns the pipeline ID of where the node is currently connected to
        :param pipeline: pipeline definition where the link is located
        :param node_id: the node ID of the node
        :return: the pipeline ID of where the node is located
        """
        pipeline_json = json.loads(json.dumps(pipeline))
        for single_pipeline in pipeline_json['pipelines']:
            node_list = single_pipeline['nodes']
            for node in node_list:
                if node['id'] == node_id:
                    return single_pipeline['id']

    async def _get_component_properties(self, pipeline_type: str, components: dict, node_op: str) -> Dict:
        """
        Retrieve the full dict of properties associated with the node_op
        :param components: list of components associated with the pipeline runtime being used e.g. kfp, airflow
        :param node_op: the node operation e.g. execute-notebook-node
        :return: a list of property names associated with the node op
        """

        if node_op == "execute-notebook-node":
            node_op = "notebooks"
        elif node_op == "execute-r-node":
            node_op = "r-script"
        elif node_op == "execute-python-node":
            node_op = "python-script"
        for category in components['categories']:
            for node_type in category['node_types']:
                if node_op == node_type['op']:
                    component: Component = \
                        await PipelineProcessorManager.instance().get_component(pipeline_type, node_op)
                    component_properties = ComponentRegistry.to_canvas_properties(component)
                    return component_properties

        return {}

    def _get_runtime_schema(self, pipeline: dict, response: ValidationResponse) -> str:
        pipeline_json = json.loads(json.dumps(pipeline))
        if not self._is_legacy_pipeline(pipeline):
            runtime = pipeline_json['pipelines'][0]['app_data']['properties'].get('runtime')
        else:
            # Assume Generic since properties field doesnt exist = older version of pipeline schema
            runtime = "Generic"

        if runtime == "Kubeflow Pipelines":
            return "kfp"
        elif runtime == "Apache Airflow":
            return "airflow"
        elif runtime == "Generic":
            return "generic"
        else:
            response.add_message(severity=ValidationSeverity.Error,
                                 message_type="invalidRuntime",
                                 message="Unsupported pipeline runtime selected in this pipeline.",
                                 data={"pipelineRuntime": runtime})

    def _get_node_names(self, pipeline: dict, node_id_list: list) -> List:
        """
        Given a node_id_list, will return the node's name for each node_id in the list, respectively
        :param pipeline: pipeline definition where the node is located
        :param node_id_list: a list of UUIDs defined in the pipeline file
        :return: a string representing the name of the node
        """
        node_name_list = []
        pipeline_json = json.loads(json.dumps(pipeline))
        for node_id in node_id_list:
            found = False
            for single_pipeline in pipeline_json['pipelines']:
                for node in single_pipeline['nodes']:
                    if node['id'] == node_id:
                        node_name_list.append(self._get_node_label(node))
                        found = True
                        break
                if found:
                    break

        return node_name_list

    def _get_node_labels(self, pipeline: dict, link_ids: List[str]) -> List[str]:
        """
        Returns the names (labels) of the nodes that are connected by
        the specified link_ids.

        :param pipeline: the pipeline dict
        :param link_ids: list of link ids from pipeline
        :return a tuple containing two node labels that are connected
        """
        if link_ids is None:
            return None

        pipeline_json = json.loads(json.dumps(pipeline))
        node_labels = []
        for link_id in link_ids:
            for single_pipeline in pipeline_json['pipelines']:
                for node in single_pipeline['nodes']:
                    if node['type'] == "execution_node":
                        for input in node.get('inputs', []):
                            for link in input.get('links', []):
                                if link['id'] == link_id:
                                    node_labels.append(self._get_node_label(node))
        return node_labels

    def _get_node_label(self, node: dict) -> str:
        """
        Returns the label for the provided node or None if the information
        cannot be derived from the inpuit dictionary.

        :param node: a dict representing a pipeline node
        :return: the label of the node
        """

        if node is None or node.get('app_data') is None:
            return None

        node_label = node['app_data'].get('label')
        if node['type'] == 'execution_node' and node['app_data'].get('ui_data'):
            node_label = node['app_data']['ui_data'].get('label')
        return node_label

    def _is_legacy_pipeline(self, pipeline: dict) -> bool:
        """
        Checks the pipeline to determine if the pipeline is an older legacy schema
        :param pipeline: the pipeline dict
        :return:
        """
        return pipeline['pipelines'][0]['app_data'].get('properties') is None

    def _is_required_property(self, property_dict: dict, node_property: str) -> bool:
        """
        Determine whether or not a component parameter is required to function correctly
        :param property_dict: the dictionary for the component
        :param node_property: the component property to check
        :return:
        """
        node_op_parameter_list = property_dict['uihints']['parameter_info']
        for parameter in node_op_parameter_list:
            if parameter['parameter_ref'] == f"elyra_{node_property}":
                return parameter['data']['required']
        return False

    def _get_component_type(self, property_dict: dict, node_property: str) -> str:
        """
        Helper function to determine the type of a node property
        :param property_dict: a dictionary containing the full list of property parameters and descriptions
        :param node_property: the property to look for
        :return: the data type associated with node_property, defaults to 'string'
        """
        for prop in property_dict['uihints']['parameter_info']:
            if prop["parameter_ref"] == f"elyra_{node_property}":
                return prop['data'].get('format', 'string')

    def _get_parent_id_list(self, pipeline_definition: PipelineDefinition,
                            node_id_list: list, parent_list: list) -> List:
        """
        Helper function to return a complete list of parent node_ids
        :param pipeline_definition: the complete pipeline definition
        :param node_id_list: list of parent node ids
        :param parent_list: the list to add additional found parent node ids
        :return:
        """
        for node_id in node_id_list:
            node = pipeline_definition.get_node(node_id)
            if node:
                if node.type in ['execution_node', 'super_node']:
                    parent_list.append(node_id)
                    node_ids = list(x.get('node_id_ref', None) for x in node.component_links)
                    for nid in node_ids:  # look-ahead to determine if node is a binding node
                        if pipeline_definition.get_node(nid).type == 'binding':
                            node_ids.remove(nid)
                            for super_node in pipeline_definition.get_supernodes():
                                if super_node['inputs'][0]['subflow_node_ref'] == nid:
                                    links = list(x.get('node_id_ref', None) for x in super_node.component_links)
                                    node_ids.append(links)
                    self._get_parent_id_list(pipeline_definition, node_ids, parent_list)
                else:  # binding node
                    pass
        return parent_list



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABC
from abc import abstractmethod
from datetime import datetime
import os
from subprocess import CalledProcessError
from subprocess import PIPE
from subprocess import run
import sys
import time
from typing import Dict
from typing import List
from typing import Optional

from jupyter_server.gateway.managers import GatewayClient
import papermill
from traitlets import log

from elyra.pipeline.component_registry import ComponentRegistry
from elyra.pipeline.pipeline import GenericOperation
from elyra.pipeline.processor import PipelineProcessor
from elyra.pipeline.processor import PipelineProcessorResponse
from elyra.util.path import get_absolute_path


class LocalPipelineProcessor(PipelineProcessor):
    """
    Local pipeline processor runs a pipeline locally. The scope of this runtime is
    simply to automate the execution of multiple notebooks from a pipeline as if they
    were being executed manually. Any additional support is out of the scope of this
    processor. If a notebook doesn't run using `run all` it will fail in the
    same way when processed by this processor. Also, any relationship or specific capabilities
    associated with a particular runtime is not supported by local mode and any additional properties
    other then the specific file to be executed is ignored by the local processor.

    Note: Execution happens in-place and a ledger of runs will be available at $TMPFILE/elyra/pipeline-name-<timestamp>
    """
    _operation_processor_registry: Dict
    _type = 'local'

    def __init__(self, root_dir, **kwargs):
        super().__init__(root_dir, **kwargs)
        notebook_op_processor = NotebookOperationProcessor(self.root_dir)
        python_op_processor = PythonScriptOperationProcessor(self.root_dir)
        r_op_processor = RScriptOperationProcessor(self.root_dir)
        self._operation_processor_registry = {
            notebook_op_processor.operation_name: notebook_op_processor,
            python_op_processor.operation_name: python_op_processor,
            r_op_processor.operation_name: r_op_processor,
        }

    @property
    def type(self):
        return self._type

    def get_components(self):
        return ComponentRegistry.get_generic_components()

    def process(self, pipeline):
        """
        Process a pipeline locally.
        The pipeline execution consists on properly ordering the operations
        based on it's dependency graph and than delegating the execution
        to proper executor (e.g. papermill to notebooks)
        """

        self.log_pipeline_info(pipeline.name, "processing pipeline")
        t0_all = time.time()

        # This unique run identifier is made available to all
        # notebooks | scripts in environment variable ELYRA_RUN_NAME
        # during pipeline execution. The id itself is not persisted
        # after the run completes.
        # The motivation is that notebooks | scripts might want to generate
        # artifact names that are unique for each run to avoid overwriting.
        # For example, a saved model file might be named
        # `my-trained-model-{ELYRA_RUN_NAME}.ext`.
        elyra_run_name = f'{pipeline.name}-{datetime.now().strftime("%m%d%H%M%S")}'

        # Sort operations based on dependency graph (topological order)
        operations = PipelineProcessor._sort_operations(pipeline.operations)
        for operation in operations:
            assert isinstance(operation, GenericOperation)
            try:
                t0 = time.time()
                operation_processor = self._operation_processor_registry[operation.classifier]
                operation_processor.process(operation, elyra_run_name)
                self.log_pipeline_info(pipeline.name, f"completed {operation.filename}",
                                       operation_name=operation.name,
                                       duration=(time.time() - t0))
            except Exception as ex:
                raise RuntimeError(f'Error processing operation {operation.name} {str(ex)}') from ex

        self.log_pipeline_info(pipeline.name, "pipeline processed", duration=(time.time() - t0_all))

        return LocalPipelineProcessorResponse()

    def export(self, pipeline, pipeline_export_format, pipeline_export_path, overwrite):
        raise NotImplementedError('Local pipelines does not support export functionality')


class LocalPipelineProcessorResponse(PipelineProcessorResponse):

    _type = 'local'

    def __init__(self):
        super().__init__('', '', '')

    @property
    def type(self):
        return self._type


class OperationProcessor(ABC):

    _operation_name: str = None

    def __init__(self):
        self.log = log.get_logger()

    @property
    def operation_name(self) -> str:
        return self._operation_name

    @abstractmethod
    def process(self, operation: GenericOperation, elyra_run_name: str):
        raise NotImplementedError

    @staticmethod
    def _collect_envs(operation: GenericOperation, elyra_run_name: str) -> Dict:
        envs = os.environ.copy()  # Make sure this process's env is "available" in the kernel subprocess
        envs.update(operation.env_vars_as_dict())
        envs['ELYRA_RUNTIME_ENV'] = "local"  # Special case
        envs['ELYRA_RUN_NAME'] = elyra_run_name
        return envs


class FileOperationProcessor(OperationProcessor):

    MAX_ERROR_LEN: int = 80

    def __init__(self, root_dir: str):
        super().__init__()
        self._root_dir = root_dir

    @property
    def operation_name(self) -> str:
        return self._operation_name

    @abstractmethod
    def process(self, operation: GenericOperation, elyra_run_name: str):
        raise NotImplementedError

    def get_valid_filepath(self, op_filename: str) -> str:
        filepath = get_absolute_path(self._root_dir, op_filename)
        if not os.path.exists(filepath):
            raise FileNotFoundError(f'Could not find {filepath}')
        if not os.path.isfile(filepath):
            raise ValueError(f'Not a file: {filepath}')
        return filepath

    def log_and_raise(self,
                      file_name: str,
                      ex: Exception,
                      data_capture_msg: Optional[str] = None) -> None:
        """Log and raise the exception that occurs when processing file_name.

        If the exception's message is longer than MAX_ERROR_LEN, it will be
        truncated with an ellipses (...) when raised.  The complete message
        will be logged.
        """
        self.log.error(f'Error executing {file_name}: {str(ex)}')
        if data_capture_msg:
            self.log.info(data_capture_msg)
        truncated_msg = FileOperationProcessor._truncate_msg(str(ex))
        raise RuntimeError(f'({file_name}): {truncated_msg}') from ex

    @staticmethod
    def _truncate_msg(msg: str) -> str:
        """Truncates the msg string to be less that MAX_ERROR_LEN.

        If msg is longer than MAX_ERROR_LEN, the first space is found from the right,
        then ellipses (...) are appended to that location so that they don't appear
        in the middle of a word.  As a result, the truncation could result in lengths
        less than the max+2.
        """
        if len(msg) < FileOperationProcessor.MAX_ERROR_LEN:
            return msg
        # locate the first whitespace from the 80th character and truncate from there
        last_space = msg.rfind(' ', 0, FileOperationProcessor.MAX_ERROR_LEN)
        if last_space >= 0:
            return msg[:last_space] + "..."
        return msg[:FileOperationProcessor.MAX_ERROR_LEN]


class NotebookOperationProcessor(FileOperationProcessor):
    _operation_name = 'execute-notebook-node'

    def process(self, operation: GenericOperation, elyra_run_name: str):
        filepath = self.get_valid_filepath(operation.filename)

        file_dir = os.path.dirname(filepath)
        file_name = os.path.basename(filepath)

        self.log.debug(f'Processing notebook: {filepath}')

        # We'll always use the ElyraEngine.  This engine is essentially the default Papermill engine
        # but allows for environment variables to be passed to the kernel process (via 'kernel_env').
        # If the current notebook server is running with Enterprise Gateway configured, we will also
        # point the 'kernel_manager_class' to GatewayKernelManager so that notebooks run as they
        # would outside of Elyra.  Current working directory (cwd) is specified both for where papermill
        # runs the notebook (cwd) and where the directory of the kernel process (kernel_cwd).  The latter
        # of which is important when EG is configured.
        additional_kwargs = dict()
        additional_kwargs['engine_name'] = "ElyraEngine"
        additional_kwargs['cwd'] = file_dir  # For local operations, papermill runs from this dir
        additional_kwargs['kernel_cwd'] = file_dir
        additional_kwargs['kernel_env'] = OperationProcessor._collect_envs(operation, elyra_run_name)
        if GatewayClient.instance().gateway_enabled:
            additional_kwargs['kernel_manager_class'] = 'jupyter_server.gateway.managers.GatewayKernelManager'

        t0 = time.time()
        try:
            papermill.execute_notebook(
                filepath,
                filepath,
                **additional_kwargs
            )
        except papermill.PapermillExecutionError as pmee:
            self.log.error(f'Error executing {file_name} in cell {pmee.exec_count}: ' +
                           f'{str(pmee.ename)} {str(pmee.evalue)}')
            raise RuntimeError(f'({file_name}) in cell {pmee.exec_count}: ' +
                               f'{str(pmee.ename)} {str(pmee.evalue)}') from pmee
        except Exception as ex:
            self.log_and_raise(file_name, ex)

        t1 = time.time()
        duration = (t1 - t0)
        self.log.debug(f'Execution of {file_name} took {duration:.3f} secs.')


class ScriptOperationProcessor(FileOperationProcessor):

    _script_type: str = None

    def get_argv(self, filepath) -> List[str]:
        raise NotImplementedError

    def process(self, operation: GenericOperation, elyra_run_name: str):
        filepath = self.get_valid_filepath(operation.filename)

        file_dir = os.path.dirname(filepath)
        file_name = os.path.basename(filepath)

        self.log.debug(f'Processing {self._script_type} script: {filepath}')

        argv = self.get_argv(filepath)
        envs = OperationProcessor._collect_envs(operation, elyra_run_name)
        data_capture_msg = f"Data capture for previous error:\n"\
                           f" command: {' '.join(argv)}\n"\
                           f" working directory: {file_dir}\n"\
                           f" environment variables: {envs}"
        t0 = time.time()
        try:
            run(argv, cwd=file_dir, env=envs, check=True, stderr=PIPE)
        except CalledProcessError as cpe:
            error_msg = str(cpe.stderr.decode())
            self.log.error(f'Error executing {file_name}: {error_msg}')
            # Log process information to aid with troubleshooting
            self.log.info(data_capture_msg)

            error_trim_index = error_msg.rfind('\n', 0, error_msg.rfind('Error'))
            if error_trim_index != -1:
                raise RuntimeError(f'({file_name}): {error_msg[error_trim_index:].strip()}') from cpe
            else:
                raise RuntimeError(f'({file_name})') from cpe
        except Exception as ex:
            self.log_and_raise(file_name, ex, data_capture_msg)

        t1 = time.time()
        duration = (t1 - t0)
        self.log.debug(f'Execution of {file_name} took {duration:.3f} secs.')


class PythonScriptOperationProcessor(ScriptOperationProcessor):
    _operation_name = 'execute-python-node'
    _script_type = 'Python'

    def get_argv(self, file_path) -> List[str]:
        return [f"{sys.executable}", file_path]


class RScriptOperationProcessor(ScriptOperationProcessor):
    _operation_name = 'execute-r-node'
    _script_type = 'R'

    def get_argv(self, file_path) -> List[str]:
        return ['Rscript', file_path]



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABC
from abc import abstractmethod
import ast
import asyncio
import functools
import os
import time
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import entrypoints
from minio.error import SignatureDoesNotMatch
from traitlets.config import Bool
from traitlets.config import LoggingConfigurable
from traitlets.config import SingletonConfigurable
from traitlets.config import Unicode
from urllib3.exceptions import MaxRetryError

from elyra.metadata.manager import MetadataManager
from elyra.pipeline.component import Component
from elyra.pipeline.component import ComponentParser
from elyra.pipeline.component_registry import ComponentRegistry
from elyra.pipeline.pipeline import GenericOperation
from elyra.pipeline.pipeline import Operation
from elyra.pipeline.pipeline import Pipeline
from elyra.util.archive import create_temp_archive
from elyra.util.cos import CosClient
from elyra.util.path import get_expanded_path

elyra_log_pipeline_info = os.getenv("ELYRA_LOG_PIPELINE_INFO", True)


class PipelineProcessorRegistry(SingletonConfigurable):
    _processors = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.root_dir = get_expanded_path(kwargs.get('root_dir'))
        # Register all known processors based on entrypoint configuration
        for processor in entrypoints.get_group_all('elyra.pipeline.processors'):
            try:
                # instantiate an actual instance of the processor
                processor_instance = processor.load()(self.root_dir, parent=kwargs.get('parent'))  # Load an instance
                self.log.info(f'Registering {processor.name} processor '
                              f'"{processor.module_name}.{processor.object_name}"...')
                self.add_processor(processor_instance)
            except Exception as err:
                # log and ignore initialization errors
                self.log.error(f'Error registering {processor.name} processor '
                               f'"{processor.module_name}.{processor.object_name}" - {err}')

    def add_processor(self, processor):
        self.log.debug(f'Registering processor {processor.type}')
        self._processors[processor.type] = processor

    def get_processor(self, processor_type: str):
        if self.is_valid_processor(processor_type):
            return self._processors[processor_type]
        else:
            raise RuntimeError('Could not find pipeline processor for [{}]'.format(processor_type))

    def is_valid_processor(self, processor_type: str) -> bool:
        return processor_type in self._processors.keys()


class PipelineProcessorManager(SingletonConfigurable):
    _registry: PipelineProcessorRegistry

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.root_dir = get_expanded_path(kwargs.get('root_dir'))
        self._registry = PipelineProcessorRegistry.instance()

    def _get_processor_for_runtime(self, processor_type: str):
        processor = self._registry.get_processor(processor_type)
        return processor

    def is_supported_runtime(self, processor_type: str) -> bool:
        return self._registry.is_valid_processor(processor_type)

    async def get_components(self, processor_type):
        processor = self._get_processor_for_runtime(processor_type)

        res = await asyncio.get_event_loop().run_in_executor(None, processor.get_components)
        return res

    async def get_component(self, processor_type, component_id):
        processor = self._get_processor_for_runtime(processor_type)

        res = await asyncio.get_event_loop().\
            run_in_executor(None, functools.partial(processor.get_component, component_id=component_id))
        return res

    async def process(self, pipeline):
        processor = self._get_processor_for_runtime(pipeline.runtime)

        res = await asyncio.get_event_loop().run_in_executor(None, processor.process, pipeline)
        return res

    async def export(self, pipeline, pipeline_export_format, pipeline_export_path, overwrite):
        processor = self._get_processor_for_runtime(pipeline.runtime)

        res = await asyncio.get_event_loop().run_in_executor(
            None, processor.export, pipeline, pipeline_export_format, pipeline_export_path, overwrite)
        return res


class PipelineProcessorResponse(ABC):

    _type = None

    def __init__(self, run_url, object_storage_url, object_storage_path):
        self._run_url = run_url
        self._object_storage_url = object_storage_url
        self._object_storage_path = object_storage_path

    @property
    @abstractmethod
    def type(self):
        raise NotImplementedError()

    @property
    def run_url(self):
        """
        :return: The runtime URL to access the pipeline experiment
        """
        return self._run_url

    @property
    def object_storage_url(self):
        """
        :return: The object storage URL to access the pipeline outputs
                 and processed notebooks
        """
        return self._object_storage_url

    @property
    def object_storage_path(self):
        """
        :return: The object storage working directory path where the pipeline outputs
                 and processed notebooks are located
        """
        return self._object_storage_path

    def to_json(self):
        return {"platform": self.type,
                "run_url": self.run_url,
                "object_storage_url": self.object_storage_url,
                "object_storage_path": self.object_storage_path
                }


class PipelineProcessor(LoggingConfigurable):  # ABC

    _type: str = None

    _component_registry: ComponentRegistry = None

    root_dir = Unicode(allow_none=True)

    enable_pipeline_info = Bool(config=True,
                                default_value=(os.getenv('ELYRA_ENABLE_PIPELINE_INFO', 'true').lower() == 'true'),
                                help="""Produces formatted logging of informational messages with durations
                                (default=True). (ELYRA_ENABLE_PIPELINE_INFO env var)""")

    def __init__(self, root_dir, **kwargs):
        super().__init__(**kwargs)
        self.root_dir = root_dir

    @property
    @abstractmethod
    def type(self) -> str:
        raise NotImplementedError()

    def get_components(self) -> List[Component]:
        """
        Retrieve components common to all runtimes
        """
        components: List[Component] = ComponentRegistry.get_generic_components()

        # Retrieve runtime-specific components
        if self._component_registry:
            components.extend(self._component_registry.get_all_components())

        return components

    def get_component(self, component_id: str) -> Optional[Component]:
        """
        Retrieve runtime-specific component details if component_id is not one of the generic set
        """

        if component_id not in ('notebook', 'python-script', 'r-script'):
            return self._component_registry.get_component(component_id=component_id)

        return ComponentRegistry.get_generic_component(component_id)

    @abstractmethod
    def process(self, pipeline) -> PipelineProcessorResponse:
        raise NotImplementedError()

    @abstractmethod
    def export(self, pipeline, pipeline_export_format, pipeline_export_path, overwrite):
        raise NotImplementedError()

    def log_pipeline_info(self, pipeline_name: str, action_clause: str, **kwargs):
        """
        Produces a formatted log INFO message used entirely for support purposes.

        This method is intended to be called for any entries that should be captured across aggregated
        log files to identify steps within a given pipeline and each of its operations.  As a result,
        calls to this method should produce single-line entries in the log (no embedded newlines).
        Each entry is prefixed with the pipeline name.  This functionality can be disabled by setting
        PipelineProcessor.enable_pipeline_info = False (or via env ELYRA_ENABLE_PIPELINE_INFO).

        General logging should NOT use this method but use logger.<level>() statements directly.

        :param pipeline_name: str representing the name of the pipeline that is being executed
        :param action_clause: str representing the action that is being logged
        :param **kwargs: dict representing the keyword arguments.  Recognized keywords include:
               operation_name: str representing the name of the operation applicable for this entry
               duration: float value representing the duration of the action being logged
        """
        if self.enable_pipeline_info:
            duration = kwargs.get('duration')
            duration_clause = f"({duration:.3f} secs)" if duration else ""

            operation_name = kwargs.get('operation_name')
            op_clause = f":'{operation_name}'" if operation_name else ""

            self.log.info(f"{self._type} '{pipeline_name}'{op_clause} - {action_clause} {duration_clause}")

    @staticmethod
    def _propagate_operation_inputs_outputs(pipeline: Pipeline, sorted_operations: List[Operation]) -> None:
        """
        All previous operation outputs should be propagated throughout the pipeline.
        In order to process this recursively, the current operation's inputs should be combined
        from its parent's inputs (which, themselves are derived from the outputs of their parent)
        and its parent's outputs.
        """
        for operation in sorted_operations:
            parent_io = set()  # gathers inputs & outputs relative to parent
            for parent_operation_id in operation.parent_operation_ids:
                parent_operation = pipeline.operations[parent_operation_id]
                if parent_operation.inputs:
                    parent_io.update(parent_operation.inputs)
                if parent_operation.outputs:
                    parent_io.update(parent_operation.outputs)

            if parent_io:
                parent_io.update(operation.inputs)
                operation.inputs = list(parent_io)

    @staticmethod
    def _sort_operations(operations_by_id: dict) -> List[Operation]:
        """
        Sort the list of operations based on its dependency graph
        """
        ordered_operations = []

        for operation in operations_by_id.values():
            PipelineProcessor._sort_operation_dependencies(operations_by_id,
                                                           ordered_operations,
                                                           operation)

        return ordered_operations

    @staticmethod
    def _sort_operation_dependencies(operations_by_id: dict, ordered_operations: list, operation: Operation) -> None:
        """
        Helper method to the main sort operation function
        """
        # Optimization: check if already processed
        if operation not in ordered_operations:
            # process each of the dependencies that needs to be executed first
            for parent_operation_id in operation.parent_operation_ids:
                parent_operation = operations_by_id[parent_operation_id]
                if parent_operation not in ordered_operations:
                    PipelineProcessor._sort_operation_dependencies(operations_by_id,
                                                                   ordered_operations,
                                                                   parent_operation)
            ordered_operations.append(operation)


class RuntimePipelineProcessor(PipelineProcessor):

    @property
    def component_parser(self) -> ComponentParser:
        return self._component_parser

    @property
    def component_registry(self) -> ComponentRegistry:
        return self._component_registry

    def __init__(self, root_dir: str, component_parser: ComponentParser, **kwargs):
        super().__init__(root_dir, **kwargs)

        self._component_parser = component_parser
        self._component_registry = ComponentRegistry(component_parser)

    def _get_dependency_archive_name(self, operation):
        artifact_name = os.path.basename(operation.filename)
        (name, ext) = os.path.splitext(artifact_name)
        return name + '-' + operation.id + ".tar.gz"

    def _get_dependency_source_dir(self, operation):
        return os.path.join(self.root_dir, os.path.dirname(operation.filename))

    def _generate_dependency_archive(self, operation):
        archive_artifact_name = self._get_dependency_archive_name(operation)
        archive_source_dir = self._get_dependency_source_dir(operation)

        dependencies = [os.path.basename(operation.filename)]
        dependencies.extend(operation.dependencies)

        archive_artifact = create_temp_archive(archive_name=archive_artifact_name,
                                               source_dir=archive_source_dir,
                                               filenames=dependencies,
                                               recursive=operation.include_subdirectories,
                                               require_complete=True)

        return archive_artifact

    def _upload_dependencies_to_object_store(self, runtime_configuration, pipeline_name, operation):
        operation_artifact_archive = self._get_dependency_archive_name(operation)
        cos_directory = pipeline_name
        # upload operation dependencies to object store
        try:
            t0 = time.time()
            dependency_archive_path = self._generate_dependency_archive(operation)
            self.log_pipeline_info(pipeline_name,
                                   f"generated dependency archive: {dependency_archive_path}",
                                   operation_name=operation.name,
                                   duration=(time.time() - t0))

            cos_client = CosClient(config=runtime_configuration)

            t0 = time.time()
            cos_client.upload_file_to_dir(dir=cos_directory,
                                          file_name=operation_artifact_archive,
                                          file_path=dependency_archive_path)
            self.log_pipeline_info(pipeline_name,
                                   f"uploaded dependency archive to: {cos_directory}/{operation_artifact_archive}",
                                   operation_name=operation.name,
                                   duration=(time.time() - t0))

        except FileNotFoundError as ex:
            self.log.error("Dependencies were not found building archive for operation: {}".
                           format(operation.name), exc_info=True)
            raise FileNotFoundError("Node '{}' referenced dependencies that were not found: {}".
                                    format(operation.name, ex)) from ex
        except MaxRetryError as ex:
            cos_endpoint = runtime_configuration.metadata.get('cos_endpoint')
            self.log.error("Connection was refused when attempting to connect to : {}".
                           format(cos_endpoint), exc_info=True)
            raise RuntimeError("Connection was refused when attempting to upload artifacts to : '{}'. Please "
                               "check your object storage settings. ".format(cos_endpoint)) from ex
        except SignatureDoesNotMatch as ex:
            raise RuntimeError("Connection was refused due to incorrect Object Storage credentials. " +
                               "Please validate your runtime configuration details and retry.") from ex
        except BaseException as ex:
            self.log.error("Error uploading artifacts to object storage for operation: {}".
                           format(operation.name), exc_info=True)
            raise ex from ex

    def _get_metadata_configuration(self, schemaspace, name=None):
        """
        Retrieve associated metadata configuration based on schemaspace provided and optional instance name
        :return: metadata in json format
        """
        try:
            if not name:
                return MetadataManager(schemaspace=schemaspace).get_all()
            else:
                return MetadataManager(schemaspace=schemaspace).get(name)
        except BaseException as err:
            self.log.error(f'Error retrieving metadata configuration for {name}', exc_info=True)
            raise RuntimeError(f'Error retrieving metadata configuration for {name}', err) from err

    def _collect_envs(self, operation: GenericOperation, **kwargs) -> Dict:
        """
        Collect the envs stored on the Operation and set the system-defined ELYRA_RUNTIME_ENV

        Note: subclasses should call their superclass (this) method first.
        :return: dictionary containing environment name/value pairs
        """

        envs: Dict = operation.env_vars_as_dict(logger=self.log)
        envs['ELYRA_RUNTIME_ENV'] = self.type

        if 'cos_secret' not in kwargs or not kwargs['cos_secret']:
            envs['AWS_ACCESS_KEY_ID'] = kwargs['cos_username']
            envs['AWS_SECRET_ACCESS_KEY'] = kwargs['cos_password']
        else:  # ensure the "access-key" envs are NOT present...
            envs.pop('AWS_ACCESS_KEY_ID', None)
            envs.pop('AWS_SECRET_ACCESS_KEY', None)

        # Convey pipeline logging enablement to operation
        envs['ELYRA_ENABLE_PIPELINE_INFO'] = str(self.enable_pipeline_info)
        return envs

    def _process_dictionary_value(self, value: str) -> Union[Dict, str]:
        """
        For component parameters of type dictionary, the user-entered string value given in the pipeline
        JSON should be converted to the appropriate Dict format, if possible. If a Dict cannot be formed,
        log and return stripped string value.
        """
        if not value:
            return {}

        value = value.strip()
        if value == "None":
            return {}

        converted_dict = None
        if value.startswith('{') and value.endswith('}'):
            try:
                converted_dict = ast.literal_eval(value)
            except Exception:
                pass

        # Value could not be successfully converted to dictionary
        if not isinstance(converted_dict, dict):
            self.log.debug(f"Could not convert entered parameter value `{value}` to dictionary")
            return value

        return converted_dict

    def _process_list_value(self, value: str) -> Union[List, str]:
        """
        For component parameters of type list, the user-entered string value given in the pipeline JSON
        should be converted to the appropriate List format, if possible. If a List cannot be formed,
        log and return stripped string value.
        """
        if not value:
            return []

        value = value.strip()
        if value == "None":
            return []

        converted_list = None
        if value.startswith('[') and value.endswith(']'):
            try:
                converted_list = ast.literal_eval(value)
            except Exception:
                pass

        # Value could not be successfully converted to list
        if not isinstance(converted_list, list):
            self.log.debug(f"Could not convert entered parameter value `{value}` to list")
            return value

        return converted_list



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABC
from abc import abstractmethod
import json
import os
from typing import List


class NodeFile(object):
    """Base class for input and output node files"""
    def __init__(self, filename: str) -> None:
        self.filename = filename


class InputNodeFile(NodeFile):
    """Given a filename, it ensures the file exists and can read its contents."""
    def __init__(self, filename: str) -> None:
        super().__init__(filename)
        self.data = None

        if not os.path.exists(self.filename):
            raise FileNotFoundError("File '{}' does not exist!".format(self.filename))

    def read(self) -> str:
        with open(self.filename) as f:
            self.data = f.read()
        return self.data

    def data(self) -> str:
        return self.data


class OutputNodeFile(NodeFile):
    """Given a filename, it ensures the file does not exist and will write data to that file."""
    def __init__(self, filename: str) -> None:
        super().__init__(filename)

        # Don't enforce output file existence here - break idempotency
        # if os.path.exists(self.filename):
        #    raise FileExistsError("File '{}' already exists!".format(self.filename))

    def write(self, data) -> None:
        with open(self.filename, 'w+') as f:
            f.write(data)


class ExecutionNode(ABC):
    """Represents an excutable node of a pipeline.  This class must be subclassed."""

    node_name = None
    filename = None
    extension = None

    def __init__(self) -> None:
        self.filename = os.getenv('NODE_FILENAME')
        if not self.filename:
            raise ValueError("NODE_FILENAME environment variable must be set!")

        node_file_splits = os.path.basename(self.filename).split(".")
        self.node_name = node_file_splits[0]
        self.extension = node_file_splits[1]
        self.validate()

    def validate(self) -> None:
        """Validate the filename as best as possible, depending on subclass. """

        # Validate its extension and that the file exists.
        self.validate_extension()
        if not os.path.exists(self.filename):
            raise FileNotFoundError("ExecutionNode filename '{}' does not exist!".format(self.filename))

    def run(self) -> None:
        self.process_inputs("INPUT_FILENAMES")
        self.perform_experiment()
        self.process_outputs("OUTPUT_FILENAMES")

    def perform_experiment(self) -> None:
        """Emulates the experiment to run."""
        print(f"NODE_NAME: {self.node_name}")

        runtime_env = os.getenv("ELYRA_RUNTIME_ENV")
        assert runtime_env == "local", "ELYRA_RUNTIME_ENV has not been set to 'local'!"
        print(f"ELYRA_RUNTIME_ENV: {runtime_env}")

        run_name = os.getenv("ELYRA_RUN_NAME")
        assert run_name is not None, "ELYRA_RUN_NAME is not set!"
        print(f"ELYRA_RUN_NAME: {run_name}")

        pipeline_name = os.getenv("PIPELINE_NAME")
        print(f"PIPELINE_NAME: {pipeline_name}")
        assert pipeline_name is not None, "PIPELINE_NAME is not set!"

        assert run_name.startswith(pipeline_name), "ELYRA_RUN_NAME does not start with pipeline name!"

    def process_inputs(self, env_var: str) -> List[InputNodeFile]:
        """Given an environment variable `env_var`, that contains a SEMI-COLON-separated
           list of filenames, it processes each entry by instantiating an instance of
           InputNodeFile corresponding to each entry and returns the list of instances.
        """
        inputs = []
        filenames = os.getenv(env_var, "").split(';')

        for filename in filenames:
            if filename:
                inputs.append(InputNodeFile(filename))

        for input_file in inputs:
            payload = json.loads(input_file.read())
            print(f"FROM: {payload.get('node')}")
            assert payload.get('run_name') == os.getenv("ELYRA_RUN_NAME")

        return inputs

    def process_outputs(self, env_var: str) -> List[OutputNodeFile]:
        """Given an environment variable `env_var`, that contains a SEMI-COLON-separated
           list of filenames, it processes each entry by instantiating an instance of
           OutputNodeFile corresponding to each entry and returns the list of instances.
        """
        outputs = []
        filenames = os.getenv(env_var, "").split(';')

        for filename in filenames:
            if filename:
                outputs.append(OutputNodeFile(filename))

        # Include ELYRA_RUN_NAME in all outputs - which are verified when used as inputs
        payload = {"node": self.node_name, "run_name": os.getenv("ELYRA_RUN_NAME")}
        for output_file in outputs:
            output_file.write(json.dumps(payload))

        return outputs

    @abstractmethod
    def expected_extension(self) -> str:
        raise NotImplementedError("Method 'expected_extension()' must be implemented by subclass '{}'!".
                                  format(self.__class__.__name__))

    def validate_extension(self) -> None:
        if self.expected_extension() != self.extension:
            raise ValueError("Filename '{}' does not have a proper extension: '{}'".
                             format(self.filename, self.expected_extension()))


class NotebookNode(ExecutionNode):
    """Represents a Notebook execution node of a pipeline."""

    def expected_extension(self) -> str:
        return "ipynb"

    def validate(self) -> None:
        """For notebooks, we can also ensure the file can be loaded as JSON."""
        super().validate()

        # Confirm file can be loaded as JSON
        with open(self.filename) as f:
            json.load(f)


class PythonNode(ExecutionNode):
    """Represents a Python file execution node of a pipeline."""

    def expected_extension(self) -> str:
        return "py"



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta
import io
import json
import os
from typing import Dict
from typing import List

import entrypoints
try:
    from kfp_tekton import TektonClient
except ImportError:
    # We may not have kfp-tekton available and that's okay!
    TektonClient = None

from elyra.metadata.schema import SchemasProvider


class ElyraSchemasProvider(SchemasProvider, metaclass=ABCMeta):
    """Base class used for retrieving Elyra-based schema files from its metadata/schemas directory."""

    def __init__(self):
        # get set of registered runtimes
        self._runtime_processor_names = set()
        for processor in entrypoints.get_group_all('elyra.pipeline.processors'):
            # load the names of the runtime processors (skip 'local')
            if processor.name == 'local':
                continue
            self._runtime_processor_names.add(processor.name)

    def get_schemas_by_name(self, schema_names: List[str]) -> List[Dict]:
        schemas = []
        schema_dir = os.path.join(os.path.dirname(__file__), 'schemas')
        schema_files = [json_file for json_file in os.listdir(schema_dir) if json_file.endswith('.json')]
        for json_file in schema_files:
            basename = os.path.splitext(json_file)[0]
            if basename in schema_names:
                schema_file = os.path.join(schema_dir, json_file)
                with io.open(schema_file, 'r', encoding='utf-8') as f:
                    schema_json = json.load(f)
                    schemas.append(schema_json)
        return schemas


class RuntimesSchemas(ElyraSchemasProvider):
    """Returns schemas relative to Runtimes schemaspace only for THIS provider."""

    elyra_processors = ['airflow', 'kfp']

    def get_schemas(self) -> List[Dict]:
        schemas = []
        kfp_needed = False
        # determine if both airflow and kfp are needed and note if kfp is needed for later
        for elyra_processor in self.elyra_processors:
            if elyra_processor in self._runtime_processor_names:
                schemas.append(elyra_processor)
                if elyra_processor == 'kfp':
                    kfp_needed = True

        runtime_schemas = self.get_schemas_by_name(schemas)
        if kfp_needed:  # Update the kfp engine enum to reflect current packages...
            # If TektonClient package is missing, navigate to the engine property
            # and remove 'tekton' entry if present and return updated result.
            if not TektonClient:
                # locate the schema and update the enum
                for schema in runtime_schemas:
                    if schema['name'] == 'kfp':
                        engine_enum: list = schema['properties']['metadata']['properties']['engine']['enum']
                        if 'Tekton' in engine_enum:
                            engine_enum.remove('Tekton')
                            schema['properties']['metadata']['properties']['engine']['enum'] = engine_enum
        return runtime_schemas


class RuntimeImagesSchemas(ElyraSchemasProvider):
    """Returns schemas relative to Runtime Images schemaspace."""
    def get_schemas(self) -> List[Dict]:
        return self.get_schemas_by_name(['runtime-image'])


class CodeSnippetsSchemas(ElyraSchemasProvider):
    """Returns schemas relative to Code Snippets schemaspace."""
    def get_schemas(self) -> List[Dict]:
        return self.get_schemas_by_name(['code-snippet'])


class ComponentRegistriesSchemas(ElyraSchemasProvider):
    """Returns schemas relative to Component Registries schemaspace."""
    def get_schemas(self) -> List[Dict]:
        schemas = self.get_schemas_by_name(['component-registry'])

        # Update runtime enum with set of currently registered runtimes
        for schema in schemas:
            schema['properties']['metadata']['properties']['runtime']['enum'] = list(self._runtime_processor_names)

        return schemas



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import OrderedDict
from datetime import datetime
import json
import os
import re
import tempfile
import time
from typing import Dict
from typing import List
from typing import Union

import autopep8
from jinja2 import Environment
from jinja2 import PackageLoader

from elyra._version import __version__
from elyra.airflow.operator import BootscriptBuilder
from elyra.metadata.schemaspaces import RuntimeImages
from elyra.metadata.schemaspaces import Runtimes
from elyra.pipeline.airflow.component_parser_airflow import AirflowComponentParser
from elyra.pipeline.pipeline import GenericOperation
from elyra.pipeline.processor import PipelineProcessor
from elyra.pipeline.processor import PipelineProcessorResponse
from elyra.pipeline.processor import RuntimePipelineProcessor
from elyra.util.git import GithubClient
from elyra.util.path import get_absolute_path


class AirflowPipelineProcessor(RuntimePipelineProcessor):
    _type = 'airflow'

    # Provide users with the ability to identify a writable directory in the
    # running container where the notebook | script is executed. The location
    # must exist and be known before the container is started.
    # Defaults to `/tmp`
    WCD = os.getenv('ELYRA_WRITABLE_CONTAINER_DIR', '/tmp').strip().rstrip('/')

    @property
    def type(self):
        return self._type

    def __init__(self, root_dir, **kwargs):
        super().__init__(root_dir, component_parser=AirflowComponentParser(), **kwargs)

    def process(self, pipeline):
        t0_all = time.time()
        timestamp = datetime.now().strftime("%m%d%H%M%S")
        pipeline_name = f'{pipeline.name}-{timestamp}'

        runtime_configuration = self._get_metadata_configuration(schemaspace=Runtimes.RUNTIMES_SCHEMASPACE_ID,
                                                                 name=pipeline.runtime_config)
        api_endpoint = runtime_configuration.metadata.get('api_endpoint')
        cos_endpoint = runtime_configuration.metadata.get('cos_endpoint')
        cos_bucket = runtime_configuration.metadata.get('cos_bucket')

        github_api_endpoint = runtime_configuration.metadata.get('github_api_endpoint')
        github_repo_token = runtime_configuration.metadata.get('github_repo_token')
        github_repo = runtime_configuration.metadata.get('github_repo')
        github_branch = runtime_configuration.metadata.get('github_branch')

        self.log_pipeline_info(pipeline_name, "Submitting pipeline")
        with tempfile.TemporaryDirectory() as temp_dir:
            pipeline_export_path = os.path.join(temp_dir, f'{pipeline_name}.py')

            self.log.debug("Creating temp directory %s", temp_dir)

            pipeline_filepath = self.create_pipeline_file(pipeline=pipeline,
                                                          pipeline_export_format="py",
                                                          pipeline_export_path=pipeline_export_path,
                                                          pipeline_name=pipeline_name)

            self.log.debug("Uploading pipeline file: %s", pipeline_filepath)

            try:
                github_client = GithubClient(server_url=github_api_endpoint,
                                             token=github_repo_token,
                                             repo=github_repo,
                                             branch=github_branch)

            except BaseException as e:
                raise RuntimeError(f'Unable to create a connection to {github_api_endpoint}: {str(e)}') from e

            github_client.upload_dag(pipeline_filepath, pipeline_name)

            self.log.info('Waiting for Airflow Scheduler to process and start the pipeline')

            github_url = github_client.get_github_url(api_url=github_api_endpoint,
                                                      repository_name=github_repo,
                                                      repository_branch=github_branch)

            self.log_pipeline_info(pipeline_name,
                                   f"pipeline pushed to git: {github_url}",
                                   duration=(time.time() - t0_all))

            return AirflowPipelineProcessorResponse(
                git_url=f'{github_url}',
                run_url=f'{api_endpoint}',
                object_storage_url=f'{cos_endpoint}',
                object_storage_path=f'/{cos_bucket}/{pipeline_name}',
            )

    def export(self, pipeline, pipeline_export_format, pipeline_export_path, overwrite):
        if pipeline_export_format not in ["py"]:
            raise ValueError("Pipeline export format {} not recognized.".format(pipeline_export_format))

        timestamp = datetime.now().strftime("%m%d%H%M%S")
        pipeline_name = f'{pipeline.name}-{timestamp}'

        absolute_pipeline_export_path = get_absolute_path(self.root_dir, pipeline_export_path)

        if os.path.exists(absolute_pipeline_export_path) and not overwrite:
            raise ValueError("File " + absolute_pipeline_export_path + " already exists.")

        self.log_pipeline_info(pipeline_name, f"exporting pipeline as a .{pipeline_export_format} file")

        new_pipeline_file_path = self.create_pipeline_file(pipeline=pipeline,
                                                           pipeline_export_format="py",
                                                           pipeline_export_path=absolute_pipeline_export_path,
                                                           pipeline_name=pipeline_name)

        return new_pipeline_file_path

    def _cc_pipeline(self, pipeline, pipeline_name):

        runtime_configuration = self._get_metadata_configuration(schemaspace=Runtimes.RUNTIMES_SCHEMASPACE_ID,
                                                                 name=pipeline.runtime_config)
        image_namespace = self._get_metadata_configuration(schemaspace=RuntimeImages.RUNTIME_IMAGES_SCHEMASPACE_ID)

        cos_endpoint = runtime_configuration.metadata['cos_endpoint']
        cos_username = runtime_configuration.metadata['cos_username']
        cos_password = runtime_configuration.metadata['cos_password']
        cos_secret = runtime_configuration.metadata.get('cos_secret')
        cos_directory = pipeline_name
        cos_bucket = runtime_configuration.metadata['cos_bucket']

        # Create dictionary that maps component Id to its ContainerOp instance
        target_ops = []

        self.log_pipeline_info(pipeline_name,
                               f"processing pipeline dependencies to: {cos_endpoint} "
                               f"bucket: {cos_bucket} folder: {pipeline_name}")

        t0_all = time.time()

        # Sort operations based on dependency graph (topological order)
        sorted_operations = PipelineProcessor._sort_operations(pipeline.operations)

        # All previous operation outputs should be propagated throughout the pipeline.
        # In order to process this recursively, the current operation's inputs should be combined
        # from its parent's inputs (which, themselves are derived from the outputs of their parent)
        # and its parent's outputs.

        PipelineProcessor._propagate_operation_inputs_outputs(pipeline, sorted_operations)

        for operation in sorted_operations:

            if isinstance(operation, GenericOperation):
                operation_artifact_archive = self._get_dependency_archive_name(operation)

                self.log.debug("Creating pipeline component:\n {op} archive : {archive}".format(
                    op=operation, archive=operation_artifact_archive))

                # Collect env variables
                pipeline_envs = self._collect_envs(operation,
                                                   cos_secret=cos_secret,
                                                   cos_username=cos_username,
                                                   cos_password=cos_password)

                # Generate unique ELYRA_RUN_NAME value and expose it as an
                # environment variable in the container.
                # Notebook | script nodes are implemented using the kubernetes_pod_operator
                # (https://airflow.apache.org/docs/apache-airflow/1.10.12/_api/airflow/contrib/operators/kubernetes_pod_operator/index.html)
                # Environment variables that are passed to this operator are
                # pre-processed by Airflow at runtime and placeholder values (expressed as '{{ xyz }}'
                #  - see https://airflow.apache.org/docs/apache-airflow/1.10.12/macros-ref#default-variables)
                # replaced.
                if pipeline_envs is None:
                    pipeline_envs = {}
                pipeline_envs['ELYRA_RUN_NAME'] = f'{pipeline_name}-{{{{ ts_nodash }}}}'

                image_pull_policy = None
                runtime_image_pull_secret = None
                for image_instance in image_namespace:
                    if image_instance.metadata['image_name'] == operation.runtime_image:
                        if image_instance.metadata.get('pull_policy'):
                            image_pull_policy = image_instance.metadata['pull_policy']
                        if image_instance.metadata.get('pull_secret'):
                            runtime_image_pull_secret = image_instance.metadata['pull_secret']
                        break

                bootscript = BootscriptBuilder(filename=operation.filename,
                                               cos_endpoint=cos_endpoint,
                                               cos_bucket=cos_bucket,
                                               cos_directory=cos_directory,
                                               cos_dependencies_archive=operation_artifact_archive,
                                               inputs=operation.inputs,
                                               outputs=operation.outputs)

                unique_operation_name = self._get_unique_operation_name(operation_name=operation.name,
                                                                        operation_list=target_ops)

                target_op = {'notebook': unique_operation_name,
                             'id': operation.id,
                             'argument_list': bootscript.container_cmd,
                             'runtime_image': operation.runtime_image,
                             'pipeline_envs': pipeline_envs,
                             'parent_operation_ids': operation.parent_operation_ids,
                             'image_pull_policy': image_pull_policy,
                             'cpu_request': operation.cpu,
                             'mem_request': operation.memory,
                             'gpu_request': operation.gpu,
                             'operator_source': operation.component_params['filename'],
                             'is_generic_operator': True
                             }

                if runtime_image_pull_secret is not None:
                    target_op['runtime_image_pull_secret'] = runtime_image_pull_secret

                target_ops.append(target_op)

                self.log_pipeline_info(pipeline_name,
                                       f"processing operation dependencies for id: {operation.id}",
                                       operation_name=operation.name)

                self._upload_dependencies_to_object_store(runtime_configuration,
                                                          pipeline_name,
                                                          operation)

            else:
                # Retrieve component from cache
                component = self._component_registry.get_component(operation.classifier)

                # Convert the user-entered value of certain properties according to their type
                for component_property in component.properties:
                    # Skip properties for which no value was given
                    if component_property.ref not in operation.component_params.keys():
                        continue

                    # Get corresponding property's value from parsed pipeline
                    property_value = operation.component_params.get(component_property.ref)

                    self.log.debug(f"Processing component parameter '{component_property.name}' "
                                   f"of type '{component_property.data_type}'")

                    if component_property.data_type == "string":
                        # Add surrounding quotation marks to string value for correct rendering
                        # in jinja DAG template
                        operation.component_params[component_property.ref] = json.dumps(property_value)
                    elif component_property.data_type == 'dictionary':
                        processed_value = self._process_dictionary_value(property_value)
                        operation.component_params[component_property.ref] = processed_value
                    elif component_property.data_type == 'list':
                        processed_value = self._process_list_value(property_value)
                        operation.component_params[component_property.ref] = processed_value

                # Remove inputs and outputs from params dict until support for data exchange is provided
                operation.component_params_as_dict.pop("inputs")
                operation.component_params_as_dict.pop("outputs")

                # Get component class from operation name
                component_class = operation.classifier.split('_')[-1]

                unique_operation_name = self._get_unique_operation_name(operation_name=operation.name,
                                                                        operation_list=target_ops)

                target_op = {'notebook': unique_operation_name,
                             'id': operation.id,
                             'module_name': component.location.rsplit('/', 1)[-1].split('.')[0],
                             'class_name': component_class,
                             'parent_operation_ids': operation.parent_operation_ids,
                             'component_params': operation.component_params_as_dict,
                             'operator_source': component.location,
                             'is_generic_operator': False
                             }
                if operation.classifier in ['spark-submit-operator', 'spark-jdbc-operator',
                                            'spark-sql-operator', 'ssh-operator']:
                    target_op['is_contrib_operator'] = True

                target_ops.append(target_op)

        ordered_target_ops = OrderedDict()

        while target_ops:
            for i in range(len(target_ops)):
                target_op = target_ops.pop(0)
                if not target_op['parent_operation_ids']:
                    ordered_target_ops[target_op['id']] = target_op
                    self.log.debug("Root Node added : %s", ordered_target_ops[target_op['id']])
                elif all(deps in ordered_target_ops.keys() for deps in target_op['parent_operation_ids']):
                    ordered_target_ops[target_op['id']] = target_op
                    self.log.debug("Dependent Node added : %s", ordered_target_ops[target_op['id']])
                else:
                    target_ops.append(target_op)

        self.log_pipeline_info(pipeline_name, "pipeline dependencies processed", duration=(time.time() - t0_all))

        return ordered_target_ops

    def create_pipeline_file(self, pipeline, pipeline_export_format, pipeline_export_path, pipeline_name):

        self.log.info('Creating pipeline definition as a .' + pipeline_export_format + ' file')
        if pipeline_export_format == "json":
            with open(pipeline_export_path, 'w', encoding='utf-8') as file:
                json.dump(pipeline_export_path, file, ensure_ascii=False, indent=4)
        else:
            # Load template from installed elyra package
            loader = PackageLoader('elyra', 'templates/airflow')
            template_env = Environment(loader=loader)

            template_env.filters['regex_replace'] = lambda string: re.sub("[-!@#$%^&*(){};:,/<>?|`~=+ ]",
                                                                          "_",
                                                                          string)  # nopep8 E731

            template = template_env.get_template('airflow_template.jinja2')

            target_ops = self._cc_pipeline(pipeline, pipeline_name)
            runtime_configuration = self._get_metadata_configuration(schemaspace=Runtimes.RUNTIMES_SCHEMASPACE_ID,
                                                                     name=pipeline.runtime_config)
            user_namespace = runtime_configuration.metadata.get('user_namespace') or 'default'
            cos_secret = runtime_configuration.metadata.get('cos_secret')

            pipeline_description = pipeline.description
            if pipeline_description is None:
                pipeline_description = f"Created with Elyra {__version__} pipeline editor using `{pipeline.source}`."

            python_output = template.render(operations_list=target_ops,
                                            pipeline_name=pipeline_name,
                                            namespace=user_namespace,
                                            cos_secret=cos_secret,
                                            kube_config_path=None,
                                            is_paused_upon_creation='False',
                                            in_cluster='True',
                                            pipeline_description=pipeline_description)

            # Write to python file and fix formatting
            with open(pipeline_export_path, "w") as fh:
                # Defer the import to postpone logger messages: https://github.com/psf/black/issues/2058
                import black
                autopep_output = autopep8.fix_code(python_output)
                output_to_file = black.format_str(autopep_output, mode=black.FileMode())
                fh.write(output_to_file)

        return pipeline_export_path

    def _get_unique_operation_name(self, operation_name: str, operation_list: list) -> str:
        unique_name_counter = 1
        unique_operation_name = operation_name
        names = [op['notebook'] for op in operation_list]
        while unique_operation_name in names:
            unique_name_counter += 1
            unique_operation_name = ''.join([operation_name, '_', str(unique_name_counter)])

        return unique_operation_name

    def _process_dictionary_value(self, value: str) -> Union[Dict, str]:
        """
        For component parameters of type dictionary, if a string value is returned from the superclass
        method, it must be converted to include surrounding quotation marks for correct rendering
        in jinja DAG template.
        """
        converted_value = super()._process_dictionary_value(value)
        if isinstance(converted_value, str):
            converted_value = json.dumps(converted_value)
        return converted_value

    def _process_list_value(self, value: str) -> Union[List, str]:
        """
        For component parameters of type list, if a string value is returned from the superclass
        method, it must be converted to include surrounding quotation marks for correct rendering
        in jinja DAG template.
        """
        converted_value = super()._process_list_value(value)
        if isinstance(converted_value, str):
            converted_value = json.dumps(converted_value)
        return converted_value


class AirflowPipelineProcessorResponse(PipelineProcessorResponse):

    _type = 'airflow'

    def __init__(self, git_url, run_url, object_storage_url, object_storage_path):
        super().__init__(run_url, object_storage_url, object_storage_path)
        self.git_url = git_url

    @property
    def type(self):
        return self._type

    def to_json(self):
        response = super().to_json()
        response['git_url'] = self.git_url
        return response



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import re
from typing import Dict
from typing import List
from typing import TypeVar

import nbformat
from traitlets.config import LoggingConfigurable

# Setup forward reference for type hint on return from class factory method.  See
# https://stackoverflow.com/questions/39205527/can-you-annotate-return-type-when-value-is-instance-of-cls/39205612#39205612
F = TypeVar('F', bound='FileReader')


class FileReader(LoggingConfigurable):
    """
    Base class for parsing a file for resources according to operation type. Subclasses set
    their own parser member variable according to their implementation language.
    """

    def __init__(self, filepath: str, **kwargs):
        super().__init__(**kwargs)
        self._filepath = filepath

    @property
    def filepath(self):
        return self._filepath

    @property
    def language(self) -> str:
        file_extension = os.path.splitext(self._filepath)[-1]
        if file_extension == '.py':
            return 'python'
        elif file_extension == '.r':
            return 'r'
        else:
            return None

    def read_next_code_chunk(self) -> List[str]:
        """
        Implements a generator for lines of code in the specified filepath. Subclasses
        may override if explicit line-by-line parsing is not feasible, e.g. with Notebooks.
        """
        with open(self._filepath) as f:
            for line in f:
                yield [line.strip()]


class NotebookReader(FileReader):
    def __init__(self, filepath: str, **kwargs):
        super().__init__(filepath, **kwargs)

        with open(self._filepath) as f:
            self._notebook = nbformat.read(f, as_version=4)
            self._language = None

            try:
                self._language = self._notebook['metadata']['kernelspec']['language'].lower()

            except KeyError:
                self.log.warning(f'No language metadata found in {self._filepath}')

    @property
    def language(self) -> str:
        return self._language

    def read_next_code_chunk(self) -> List[str]:
        for cell in self._notebook.cells:
            if cell.source and cell.cell_type == "code":
                yield cell.source.split('\n')


class ScriptParser:
    """
    Base class for parsing individual lines of code. Subclasses implement a search_expressions()
    function that returns language-specific regexes to match against code lines.
    """

    _comment_char = "#"

    def _get_line_without_comments(self, line):
        if self._comment_char in line:
            index = line.find(self._comment_char)
            line = line[:index]
        return line.strip()

    def parse_environment_variables(self, line):
        # Parse a line fed from file and match each regex in regex dictionary
        line = self._get_line_without_comments(line)
        if not line:
            return []

        matches = []
        for key, value in self.search_expressions().items():
            for pattern in value:
                regex = re.compile(pattern)
                for match in regex.finditer(line):
                    matches.append((key, match))
        return matches


class PythonScriptParser(ScriptParser):

    def search_expressions(self) -> Dict[str, List]:
        # TODO: add more key:list-of-regex pairs to parse for additional resources
        regex_dict = dict()

        # First regex matches envvar assignments of form os.environ["name"] = value w or w/o value provided
        # Second regex matches envvar assignments that use os.getenv("name", "value") with ow w/o default provided
        # Third regex matches envvar assignments that use os.environ.get("name", "value") with or w/o default provided
        # Both name and value are captured if possible
        envs = [r"os\.environ\[[\"']([a-zA-Z_]+[A-Za-z0-9_]*)[\"']\](?:\s*=(?:\s*[\"'](.[^\"']*)?[\"'])?)*",
                r"os\.getenv\([\"']([a-zA-Z_]+[A-Za-z0-9_]*)[\"'](?:\s*\,\s*[\"'](.[^\"']*)?[\"'])?",
                r"os\.environ\.get\([\"']([a-zA-Z_]+[A-Za-z0-9_]*)[\"'](?:\s*\,(?:\s*[\"'](.[^\"']*)?[\"'])?)*"]
        regex_dict["env_vars"] = envs
        return regex_dict


class RScriptParser(ScriptParser):

    def search_expressions(self) -> Dict[str, List]:
        # TODO: add more key:list-of-regex pairs to parse for additional resources
        regex_dict = dict()

        # Tests for matches of the form Sys.setenv("key" = "value")
        envs = [r"Sys\.setenv\([\"']*([a-zA-Z_]+[A-Za-z0-9_]*)[\"']*\s*=\s*[\"']*(.[^\"']*)?[\"']*\)",
                r"Sys\.getenv\([\"']*([a-zA-Z_]+[A-Za-z0-9_]*)[\"']*\)(.)*"]
        regex_dict["env_vars"] = envs
        return regex_dict


class ContentParser(LoggingConfigurable):
    parsers = {
        'python': PythonScriptParser(),
        'r': RScriptParser()
    }

    def parse(self, filepath: str) -> dict:
        """Returns a model dictionary of all the regex matches for each key in the regex dictionary"""

        properties = {"env_vars": {}, "inputs": [], "outputs": []}
        reader = self._get_reader(filepath)
        parser = self._get_parser(reader.language)

        if not parser:
            return properties

        for chunk in reader.read_next_code_chunk():
            if chunk:
                for line in chunk:
                    matches = parser.parse_environment_variables(line)
                    for key, match in matches:
                        if key == "env_vars":
                            properties[key][match.group(1)] = match.group(2)
                        else:
                            properties[key].append(match.group(1))

        return properties

    def _validate_file(self, filepath: str):
        """
        Validate file exists and is file (e.g. not a directory)
        """
        if not os.path.exists(filepath):
            raise FileNotFoundError(f'No such file or directory: {filepath}')
        if not os.path.isfile(filepath):
            raise IsADirectoryError(f'Is a directory: {filepath}')

    def _get_reader(self, filepath: str):
        """
        Find the proper reader based on the file extension
        """
        file_extension = os.path.splitext(filepath)[-1]

        self._validate_file(filepath)

        if file_extension == '.ipynb':
            return NotebookReader(filepath)
        elif file_extension in ['.py', '.r']:
            return FileReader(filepath)
        else:
            raise ValueError(f'File type {file_extension} is not supported.')

    def _get_parser(self, language: str):
        """
        Find the proper parser based on content language
        """
        parser = None
        if language:
            parser = self.parsers.get(language)

            if not parser:
                self.log.warning(f'Content parser for {language} is not available.')
        return parser



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import time
from types import SimpleNamespace
from typing import Dict
from typing import List
from typing import Optional

from jinja2 import Environment
from jinja2 import PackageLoader
from traitlets.config import LoggingConfigurable

from elyra.metadata.manager import MetadataManager
from elyra.metadata.schemaspaces import ComponentRegistries
from elyra.pipeline.component import Component
from elyra.pipeline.component import ComponentParser
from elyra.pipeline.component import ComponentReader
from elyra.pipeline.component import DirectoryComponentReader
from elyra.pipeline.component import FilesystemComponentReader
from elyra.pipeline.component import UrlComponentReader


class ComponentRegistry(LoggingConfigurable):
    """
    Component Registry, responsible to provide a list of available components
    for each runtime. The registry uses component parser to read and parse each
    component entry from the catalog and transform them into a component value object.
    """
    _cached_components: Dict[str, Component] = {}
    _cache_last_updated = None

    _generic_category_label = "Elyra"
    _generic_components: Dict[str, Component] = {
        "notebook": Component(id="notebook",
                              name="Notebook",
                              description="Run notebook file",
                              op="execute-notebook-node",
                              location_type="elyra",
                              location="elyra",
                              extensions=[".ipynb"],
                              categories=[_generic_category_label]),
        "python-script": Component(id="python-script",
                                   name="Python Script",
                                   description="Run Python script",
                                   op="execute-python-node",
                                   location_type="elyra",
                                   location="elyra",
                                   extensions=[".py"],
                                   categories=[_generic_category_label]),
        "r-script": Component(id="r-script",
                              name="R Script",
                              description="Run R script",
                              op="execute-r-node",
                              location_type="elyra",
                              location="elyra",
                              extensions=[".r"],
                              categories=[_generic_category_label])}

    def __init__(self, parser: ComponentParser, caching_enabled: bool = True, cache_ttl_in_seconds: int = 60, **kwargs):
        super().__init__(**kwargs)
        self._parser = parser

        # Initialize the cache
        self.caching_enabled = caching_enabled
        if self.caching_enabled:
            self.cache_ttl_in_seconds = cache_ttl_in_seconds
            self.update_cache()

    def get_all_components(self) -> List[Component]:
        """
        Retrieve all components; use the component registry cache if enabled
        """
        if self.caching_enabled:
            if self._is_cache_expired():
                self.update_cache()
            return list(self._cached_components.values())

        return list(self._read_component_registries().values())

    def get_component(self, component_id: str) -> Optional[Component]:
        """
        Retrieve the component with a given component_id; use component registry
        cache if enabled
        """
        component: Component
        if self.caching_enabled:
            if self._is_cache_expired():
                self.update_cache()
            component = self._cached_components.get(component_id)
        else:
            component = self._read_component_registries().get(component_id)

        if component is None:
            self.log.error(f"Component with ID '{component_id}' could not be found in any "
                           f"{self._parser.component_platform} registries.")

        return component

    def update_cache(self):
        self._cached_components = self._read_component_registries()
        self._cache_last_updated = time.time()

    def _is_cache_expired(self) -> bool:
        is_expired = True
        if self._cache_last_updated:
            now = time.time()
            elapsed = int(now - self._cache_last_updated)
            if elapsed < self.cache_ttl_in_seconds:
                is_expired = False

        return is_expired

    @staticmethod
    def get_generic_components() -> List[Component]:
        return list(ComponentRegistry._generic_components.values())

    @staticmethod
    def get_generic_component(component_id: str) -> Component:
        return ComponentRegistry._generic_components.get(component_id)

    @staticmethod
    def to_canvas_palette(components: List[Component]) -> Dict:
        """
        Converts registry components into appropriate canvas palette format
        """
        # Load jinja2 template
        loader = PackageLoader('elyra', 'templates/components')
        template_env = Environment(loader=loader)
        template = template_env.get_template('canvas_palette_template.jinja2')

        # Define a fallback category for components with no given categories
        fallback_category_name = "No Category"

        # Convert the list of all components into a dictionary of
        # component lists keyed by category
        category_dict: Dict[str, List[Component]] = {}
        for component in components:
            categories = component.categories

            # Assign a fallback category so that component is not
            # lost during palette render
            if not categories:
                categories = [fallback_category_name]

            for category in categories:
                if category not in category_dict.keys():
                    category_dict[category] = []

                if component.id not in [comp.id for comp in category_dict[category]]:
                    category_dict[category].append(component)

        # Reorder the dictionary such that components with
        # no category to render last
        fallback_category = category_dict.pop(fallback_category_name, None)
        if fallback_category:
            category_dict[fallback_category_name] = fallback_category

        # Render template
        canvas_palette = template.render(category_dict=category_dict)
        palette_json = json.loads(canvas_palette)
        return palette_json

    @staticmethod
    def to_canvas_properties(component: Component) -> Dict:
        """
        Converts registry components into appropriate canvas properties format
        """
        loader = PackageLoader('elyra', 'templates/components')
        template_env = Environment(loader=loader)

        # If component_id is one of the generic set, render with generic template,
        # else render with the runtime-specific property template
        if component.id in ('notebook', 'python-script', 'r-script'):
            template = template_env.get_template('generic_properties_template.jinja2')
        else:
            template = template_env.get_template('canvas_properties_template.jinja2')

        canvas_properties = template.render(component=component)
        properties_json = json.loads(canvas_properties)
        return properties_json

    def _read_component_registries(self) -> Dict[str, Component]:
        """
        Read through component registries and return a dictionary of components indexed by component_id.
        """
        component_dict: Dict[str, Component] = {}

        runtime_registries = self._get_registries_for_runtime()
        for registry in runtime_registries:
            self.log.debug(f"Component registry: processing components in registry '{registry['display_name']}'")

            registry_categories = registry['metadata'].get("categories", [])
            registry_location_type = registry['metadata']['location_type'].lower()

            # Assign reader based on the location type of the registry (file, directory, url)
            reader = self._get_reader(registry_location_type, self._parser.file_types)

            # Get content of component definition file for each component in this registry
            component_definitions = reader.read_component_definitions(registry['metadata']['paths'])
            for path, component_definition in component_definitions.items():

                component_entry = {
                    "location_type": reader.resource_type,
                    "location": path,
                    "categories": registry_categories,
                    "component_definition": component_definition
                }

                # Parse the component entry to get a fully qualified Component object
                components = self._parser.parse(SimpleNamespace(**component_entry)) or []
                for component in components:
                    component_dict[component.id] = component

        return component_dict

    def _get_registries_for_runtime(self) -> List[Dict]:
        """
        Retrieve the registries relevant to the calling processor instance
        """
        runtime_registries = []
        try:
            metadata_manager = MetadataManager(schemaspace=ComponentRegistries.COMPONENT_REGISTRIES_SCHEMASPACE_ID)
            all_registries = [r.to_dict(trim=True) for r in metadata_manager.get_all()]

            # Filter registries according to processor type
            runtime_registries = list(
                filter(lambda r: r['metadata']['runtime'] == self._parser.component_platform, all_registries)
            )
        except Exception:
            self.log.error(f"Could not access registries for processor: {self._parser._component_platform}")

        return runtime_registries

    def _get_reader(self, registry_location_type: str, file_types: List[str]) -> ComponentReader:
        """
        Find the proper reader based on the given registry location type
        """
        readers = {
            FilesystemComponentReader.location_type: FilesystemComponentReader(file_types),
            DirectoryComponentReader.location_type: DirectoryComponentReader(file_types),
            UrlComponentReader.location_type: UrlComponentReader(file_types)
        }

        reader = readers.get(registry_location_type)
        if not reader:
            raise ValueError(f"Unsupported registry type: '{registry_location_type}'")

        return reader



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import fnmatch
import os
import tarfile
import tempfile


WILDCARDS = ['*', '?', '[']


def create_project_temp_dir():
    temp_dir = tempfile.gettempdir()
    project_temp_dir = os.path.join(temp_dir, 'elyra')
    if not os.path.exists(project_temp_dir):
        os.mkdir(project_temp_dir)
    return project_temp_dir


def directory_in_list(directory, filenames):
    """Checks if any entries in the filenames list starts with the given directory."""
    return any(name.startswith(directory + os.sep) or fnmatch.fnmatch(directory, name) for name in filenames)


def has_wildcards(filename):
    """Returns True if the filename contains wildcard characters per https://docs.python.org/3/library/fnmatch.html """
    return len(set(WILDCARDS) & set(list(filename))) > 0


def directory_prefixed(filename):
    """Returns True if filename is prefixed by a directory (i.e., in a sub-directory."""
    return os.sep in filename and not filename.startswith(os.sep) and not filename.endswith(os.sep)


def create_temp_archive(archive_name, source_dir, filenames=None, recursive=False, require_complete=False):
    """
    Create archive file with specified list of files
    :param archive_name: the name of the archive to be created
    :param source_dir: the root folder containing source files
    :param filenames: the list of filenames, each of which can contain wildcards and/or specify subdirectories
    :param recursive: flag to include sub directories recursively
    :param require_complete: flag to indicate an exception should be raise if all filenames are not included
    :return: full path of the created archive
    """

    def tar_filter(tarinfo):
        """Filter files from the generated archive"""
        if tarinfo.type == tarfile.DIRTYPE:
            # ignore hidden directories (e.g. ipynb checkpoints and/or trash contents)
            if any(dir.startswith('.') for dir in tarinfo.name.split('/')):
                return None
            # always return the base directory (empty string) otherwise tar will be empty
            elif not tarinfo.name:
                return tarinfo
            # only include subdirectories if enabled in common properties
            elif recursive:
                return tarinfo
            # We have a directory, check if any filenames start with this value and
            # allow if found - except if a single '*' is listed (i.e., include_all) in
            # which case we don't want to add this directory since recursive is False.
            # This occurs with filenames like `data/util.py` or `data/*.py`.
            elif not include_all and directory_in_list(tarinfo.name, filenames_set):
                return tarinfo
            return None

        # We have a file at this point...

        # Special case for single wildcard entries ('*')
        if include_all:
            return tarinfo

        # Process filename
        for filename in filenames_set:
            if not filename or filename in processed_filenames:  # Skip processing
                continue

            # Match filename against candidate filename - handling wildcards
            if fnmatch.fnmatch(tarinfo.name, filename):
                # if this is a direct match, record that its been processed
                if not has_wildcards(filename) and not recursive:
                    processed_filenames.append(filename)
                matched_set.add(filename)
                return tarinfo

            # If the filename is a "flat" wildcarded value (i.e., isn't prefixed with a directory name)
            # then we should take the basename of the candidate file to perform the match against.  This
            # occurs for dependencies like *.py when include-subdirectories is enabled.
            if not directory_prefixed(filename) and has_wildcards(filename):
                if fnmatch.fnmatch(os.path.basename(tarinfo.name), filename):
                    matched_set.add(filename)
                    return tarinfo
        return None

    # Since filenames is essentially static, convert to set immediately and use the set
    filenames_set = set(filenames or [])

    # If there's a '*' - less things to check.
    include_all = len({WILDCARDS[0]} & filenames_set) > 0
    processed_filenames = []
    matched_set = set()
    temp_dir = create_project_temp_dir()
    archive = os.path.join(temp_dir, archive_name)

    with tarfile.open(archive, "w:gz", dereference=True) as tar:
        tar.add(source_dir, arcname="", filter=tar_filter)

    if require_complete and not include_all:
        # compare matched_set against filenames_set to ensure they're the same.
        if len(filenames_set) > len(matched_set):
            raise FileNotFoundError(filenames_set - matched_set)  # Only include the missing filenames

    return archive



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from typing import List
from typing import Optional

from elyra._version import __version__

"""
This helper builder constructs the bootstrapping arguments to be used as the driver for elyra's generic components
in Apache Airflow
"""

# Inputs and Outputs separator character.  If updated,
# same-named variable in bootstrapper.py must be updated!
INOUT_SEPARATOR = ';'

ELYRA_GITHUB_ORG = os.getenv("ELYRA_GITHUB_ORG", "elyra-ai")
ELYRA_GITHUB_BRANCH = os.getenv("ELYRA_GITHUB_BRANCH", "master" if 'dev' in __version__ else "v" + __version__)

ELYRA_BOOTSCRIPT_URL = os.getenv('ELYRA_BOOTSTRAP_SCRIPT_URL',
                                 'https://raw.githubusercontent.com/{org}/elyra/{branch}/elyra/airflow/bootstrapper.py'.
                                 format(org=ELYRA_GITHUB_ORG,
                                        branch=ELYRA_GITHUB_BRANCH))

ELYRA_REQUIREMENTS_URL = os.getenv('ELYRA_REQUIREMENTS_URL',
                                   'https://raw.githubusercontent.com/{org}/'
                                   'elyra/{branch}/etc/generic/requirements-elyra.txt'.
                                   format(org=ELYRA_GITHUB_ORG,
                                          branch=ELYRA_GITHUB_BRANCH))


class BootscriptBuilder(object):

    def __init__(self,
                 filename: str,
                 cos_endpoint: str,
                 cos_bucket: str,
                 cos_directory: str,
                 cos_dependencies_archive: str,
                 inputs: Optional[List[str]] = None,
                 outputs: Optional[List[str]] = None):
        """
        This helper builder constructs the bootstrapping arguments to be used as the driver for
        elyra's generic components in Apache Airflow
        :param filename: name of the file to execute
        :param :cos_endpoint: object storage endpoint e.g weaikish1.fyre.ibm.com:30442
        :param :cos_bucket: bucket to retrieve archive from
        :param :cos_directory: name of the directory in the object storage bucket to pull
        :param :cos_dependencies_archive: archive file name to get from object storage bucket e.g archive1.tar.gz
        :param inputs: comma delimited list of files to be consumed/are required by the filename
        :param outputs: comma delimited list of files produced by the filename
        """
        self.arguments = []
        self.cos_endpoint = cos_endpoint
        self.cos_bucket = cos_bucket
        self.cos_directory = cos_directory
        self.cos_dependencies_archive = cos_dependencies_archive
        self.filename = filename
        self.outputs = outputs
        self.inputs = inputs
        self.container_work_dir_root_path = "./"
        self.container_work_dir_name = "jupyter-work-dir/"
        self.container_work_dir = self.container_work_dir_root_path + self.container_work_dir_name

        if not filename:
            raise ValueError("You need to provide a filename for the operation.")

    @property
    def container_cmd(self):
        self.arguments = [f"mkdir -p {self.container_work_dir} && cd {self.container_work_dir} && "
                          f"curl -H 'Cache-Control: no-cache' -L {ELYRA_BOOTSCRIPT_URL} --output bootstrapper.py && "
                          f"curl -H 'Cache-Control: no-cache' -L {ELYRA_REQUIREMENTS_URL} "
                          f"--output requirements-elyra.txt && "
                          "python3 -m pip install packaging && "
                          "python3 -m pip freeze > requirements-current.txt && "
                          "python3 bootstrapper.py "
                          f"--cos-endpoint {self.cos_endpoint} "
                          f"--cos-bucket {self.cos_bucket} "
                          f"--cos-directory '{self.cos_directory}' "
                          f"--cos-dependencies-archive '{self.cos_dependencies_archive}' "
                          f"--file '{self.filename}' "]

        if self.inputs:
            inputs_str = self._artifact_list_to_str(self.inputs)
            self.arguments.append("--inputs '{}' ".format(inputs_str))

        if self.outputs:
            outputs_str = self._artifact_list_to_str(self.outputs)
            self.arguments.append("--outputs '{}' ".format(outputs_str))

        argument_string = ''.join(self.arguments)

        return argument_string

    def _artifact_list_to_str(self, pipeline_array):
        trimmed_artifact_list = []
        for artifact_name in pipeline_array:
            if INOUT_SEPARATOR in artifact_name:  # if INOUT_SEPARATOR is in name, throw since this is our separator
                raise \
                    ValueError("Illegal character ({}) found in filename '{}'.".format(INOUT_SEPARATOR, artifact_name))
            trimmed_artifact_list.append(artifact_name.strip())
        return INOUT_SEPARATOR.join(trimmed_artifact_list)



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os

import pytest

from elyra.contents.parser import ContentParser


def parse(filepath):
    absolute_filepath = os.path.join(os.path.dirname(__file__), filepath)

    parser = ContentParser()
    return parser.parse(absolute_filepath)


def _get_variable_names(properties):
    return list(properties['env_vars'].keys())


def test_python_notebook():
    expected_variable_names = ['VAR1', 'VAR2', 'VAR3', 'VAR4', 'VAR5', 'VAR6', 'VAR7', 'VAR8']
    properties = parse("resources/parse_python.ipynb")

    variable_names = _get_variable_names(properties)
    assert variable_names == expected_variable_names


def test_r_notebook():
    expected_variable_names = ['VAR1', 'VAR2', 'VAR3', 'VAR4', 'VAR5', 'VAR6']
    properties = parse("resources/parse_r.ipynb")

    variable_names = _get_variable_names(properties)
    assert variable_names == expected_variable_names


def test_python_script():
    expected_variable_names = ['VAR1', 'VAR2', 'VAR3', 'VAR4', 'VAR5', 'VAR6', 'VAR7', 'VAR8']
    properties = parse("resources/parse.py")

    variable_names = _get_variable_names(properties)
    assert variable_names == expected_variable_names


def test_r_script():
    expected_variable_names = ['VAR1', 'VAR2', 'VAR3', 'VAR4', 'VAR5', 'VAR6']
    properties = parse("resources/parse.r")

    variable_names = _get_variable_names(properties)
    assert variable_names == expected_variable_names


def test_empty_python_notebook():
    expected_variable_names = []
    properties = parse("resources/parse_python_empty.ipynb")

    variable_names = _get_variable_names(properties)
    assert variable_names == expected_variable_names


def test_empty_r_notebook():
    expected_variable_names = []
    properties = parse("resources/parse_r_empty.ipynb")

    variable_names = _get_variable_names(properties)
    assert variable_names == expected_variable_names


def test_empty_python_script():
    expected_variable_names = []
    properties = parse("resources/parse_empty.py")

    variable_names = _get_variable_names(properties)
    assert variable_names == expected_variable_names


def test_empty_r_script():
    expected_variable_names = []
    properties = parse("resources/parse_empty.r")

    variable_names = _get_variable_names(properties)
    assert variable_names == expected_variable_names


def test_file_not_found():
    with pytest.raises(FileNotFoundError) as e:
        parse("resources/none.py")
    assert "No such file or directory" in str(e.value)


def test_file_is_not_directory():
    directory = "dir.py"
    dir_path = os.path.join(os.path.dirname(__file__), directory)
    os.mkdir(dir_path)

    with pytest.raises(IsADirectoryError) as e:
        parse(dir_path)
    assert "Is a directory" in str(e.value)

    os.rmdir(dir_path)


def test_no_kernel():
    expected_variable_names = []
    properties = parse("resources/parse_no_kernel.ipynb")

    variable_names = _get_variable_names(properties)
    assert variable_names == expected_variable_names


def test_parser_not_set():
    expected_variable_names = []
    properties = parse("resources/parse_no_language.ipynb")

    variable_names = _get_variable_names(properties)
    assert variable_names == expected_variable_names


def test_unsupported_file_type():
    with pytest.raises(ValueError) as e:
        parse("resources/parse.txt")
    assert "File type" and "is not supported" in str(e.value)



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults


class TestOperator(BaseOperator):
    r"""
    Execute a test script.

    :param test_string_no_default: The test command description
    :type test_string_no_default: str
    :param test_bool_default: The test command bool description
    :type test_bool_default: bool
    :param test_int_default: The test command int description
    :type test_int_default: int
    :param test_dict_default: The test command dict description
    :type test_dict_default: dict
    :param test_list_default: The test command list description
    :type test_list_default: list
    :param test_string_default_value: The test command description
    :type test_string_default_value: str
    :param test_string_default_empty: The test command description
    :type test_string_default_empty: str
    :param test_bool_false: The test command bool description
    :type test_bool_false: bool
    :param test_bool_true: The test command bool description
    :type test_bool_true: bool
    :param test_int_zero: The test command int description
    :type test_int_zero: int
    :param test_int_non_zero: The test command int description
    :type test_int_non_zero: int
    :param test_unusual_type_dict: The test command description
    :type test_unusual_type_dict: a dictionary of arrays
    :param test_unusual_type_list: The test command description
    :type test_unusual_type_list: a list of strings
    :param test_unusual_type_string: The test command description
    :type test_unusual_type_string: a string
    :param test_unusual_type_notgiven: The test command description
    """

    @apply_defaults
    def __init__(self,
                 test_string_no_default,
                 test_bool_default,
                 test_int_default,
                 test_dict_default,
                 test_list_default,
                 test_string_default_value='default',
                 test_string_default_empty=None,
                 test_bool_false=False,
                 test_bool_true=True,
                 test_int_zero=0,
                 test_int_non_zero=1,
                 test_unusual_type_dict=None,
                 test_unusual_type_list=None,
                 test_unusual_type_string="",
                 test_unusual_type_notgiven="",
                 *args, **kwargs):

        super().__init__(*args, **kwargs)



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from pathlib import Path
import re
import tempfile
from unittest import mock

import github
import pytest

from elyra.metadata.metadata import Metadata
from elyra.pipeline.airflow.processor_airflow import AirflowPipelineProcessor
from elyra.pipeline.parser import PipelineParser
from elyra.pipeline.pipeline import GenericOperation
from elyra.tests.pipeline.test_pipeline_parser import _read_pipeline_resource
from elyra.util import git

PIPELINE_FILE_COMPLEX = 'resources/sample_pipelines/pipeline_dependency_complex.json'
PIPELINE_FILE_CUSTOM_COMPONENTS = 'resources/sample_pipelines/pipeline_with_airflow_components.json'


@pytest.fixture
def processor(setup_factory_data):
    processor = AirflowPipelineProcessor(os.getcwd())
    return processor


@pytest.fixture
def parsed_pipeline(request):
    pipeline_resource = _read_pipeline_resource(request.param)
    return PipelineParser().parse(pipeline_json=pipeline_resource)


@pytest.fixture
def sample_metadata():
    return {"github_api_endpoint": "https://api.github.com",
            "github_repo": "examples/test-repo",
            "github_repo_token": "c564d8dfgdsfgdfgdsfgdsfgdfsgdfsgdsfgdsfg",
            "github_branch": "main",
            "user_namespace": "default",
            "api_endpoint": "http://examples.com:31737",
            "cos_endpoint": "http://examples.com:31671",
            "cos_username": "example",
            "cos_password": "example123456",
            "cos_bucket": "test",
            "tags": []
            }


@pytest.fixture
def sample_image_metadata():
    image_one = {"image_name": 'tensorflow/tensorflow:2.0.0-py3',
                 "pull_policy": 'IfNotPresent',
                 "tags": []
                 }
    image_two = {"image_name": 'elyra/examples:1.0.0-py3',
                 "pull_policy": 'Always',
                 "tags": []
                 }

    mocked_runtime_images = [Metadata(name="test-image-metadata",
                                      display_name="test-image",
                                      schema_name="airflow",
                                      metadata=image_one
                                      ),
                             Metadata(name="test-image-metadata",
                                      display_name="test-image",
                                      schema_name="airflow",
                                      metadata=image_two
                                      )
                             ]

    return mocked_runtime_images


@pytest.fixture
def parsed_ordered_dict(monkeypatch, processor, parsed_pipeline,
                        sample_metadata, sample_image_metadata):

    mocked_runtime = Metadata(name="test-metadata",
                              display_name="test",
                              schema_name="airflow",
                              metadata=sample_metadata
                              )

    mocked_func = mock.Mock(return_value="default", side_effect=[mocked_runtime, sample_image_metadata])

    monkeypatch.setattr(processor, "_get_metadata_configuration", mocked_func)
    monkeypatch.setattr(processor, "_upload_dependencies_to_object_store", lambda x, y, z: True)
    monkeypatch.setattr(processor, "_get_dependency_archive_name", lambda x: True)

    return processor._cc_pipeline(parsed_pipeline, pipeline_name="some-name")


def read_key_pair(key_pair, sep='='):
    return {'key': key_pair.split(sep)[0].strip('" '),
            'value': key_pair.split(sep)[1].rstrip(',').strip('" ')}


def string_to_list(stringed_list):
    return stringed_list.replace(" ", "").replace("\"", "").strip('[]').split(',')


def test_processor_type(processor):
    assert processor.type == "airflow"


def test_fail_processor_type(processor):
    with pytest.raises(Exception):
        assert processor.type == "kfp"


@pytest.mark.parametrize('parsed_pipeline', [PIPELINE_FILE_COMPLEX], indirect=True)
def test_pipeline_process(monkeypatch, processor, parsed_pipeline, sample_metadata):

    mocked_runtime = Metadata(name="test-metadata",
                              display_name="test",
                              schema_name="airflow",
                              metadata=sample_metadata
                              )
    mocked_path = "/some-placeholder"

    monkeypatch.setattr(processor, "_get_metadata_configuration", lambda schemaspace, name: mocked_runtime)
    monkeypatch.setattr(processor, "create_pipeline_file",
                        lambda pipeline, pipeline_export_format, pipeline_export_path, pipeline_name: mocked_path)

    monkeypatch.setattr(github.Github, "get_repo", lambda x, y: True)
    monkeypatch.setattr(git.GithubClient, "upload_dag", lambda x, y, z: True)

    response = processor.process(pipeline=parsed_pipeline)

    assert response.run_url == sample_metadata['api_endpoint']
    assert response.object_storage_url == sample_metadata['cos_endpoint']
    # Verifies that only this substring is in the storage path since a timestamp is injected into the name
    assert "/" + sample_metadata['cos_bucket'] + "/" + "untitled" in response.object_storage_path


@pytest.mark.parametrize('parsed_pipeline', [PIPELINE_FILE_COMPLEX], indirect=True)
def test_create_file(monkeypatch, processor, parsed_pipeline, parsed_ordered_dict, sample_metadata):
    pipeline_json = _read_pipeline_resource(PIPELINE_FILE_COMPLEX)

    export_pipeline_name = "some-name"
    export_file_type = "py"

    mocked_runtime = Metadata(name="test-metadata",
                              display_name="test",
                              schema_name="airflow",
                              metadata=sample_metadata
                              )

    monkeypatch.setattr(processor, "_get_metadata_configuration", lambda name=None, schemaspace=None: mocked_runtime)
    monkeypatch.setattr(processor, "_upload_dependencies_to_object_store", lambda x, y, z: True)
    monkeypatch.setattr(processor, "_cc_pipeline", lambda x, y: parsed_ordered_dict)

    with tempfile.TemporaryDirectory() as temp_dir:
        export_pipeline_output_path = os.path.join(temp_dir, f'{export_pipeline_name}.py')

        response = processor.create_pipeline_file(parsed_pipeline,
                                                  pipeline_export_format=export_file_type,
                                                  pipeline_export_path=export_pipeline_output_path,
                                                  pipeline_name=export_pipeline_name)

        assert export_pipeline_output_path == response
        assert os.path.isfile(export_pipeline_output_path)

        file_as_lines = open(response).read().splitlines()

        assert "from airflow.contrib.operators.kubernetes_pod_operator import KubernetesPodOperator" in file_as_lines

        # Check DAG project name
        for i in range(len(file_as_lines)):
            if "args = {" == file_as_lines[i]:
                assert "project_id" == read_key_pair(file_as_lines[i + 1], sep=':')['key']
                assert export_pipeline_name == read_key_pair(file_as_lines[i + 1], sep=':')['value']

        # For every node in the original pipeline json
        for node in pipeline_json['pipelines'][0]['nodes']:
            component_parameters = node['app_data']['component_parameters']
            for i in range(len(file_as_lines)):
                # Matches a generic op with a node ID
                if f"op_{node['id'].replace('-', '_')} = KubernetesPodOperator(" in file_as_lines[i]:
                    sub_list_line_counter = 0
                    # Gets sub-list slice starting where the Notebook Op starts
                    for line in file_as_lines[i + 1:]:
                        if 'namespace=' in line:
                            assert sample_metadata['user_namespace'] == read_key_pair(line)['value']
                        elif 'cos_endpoint=' in line:
                            assert sample_metadata['cos_endpoint'] == read_key_pair(line)['value']
                        elif 'cos_bucket=' in line:
                            assert sample_metadata['cos_bucket'] == read_key_pair(line)['value']
                        elif 'name=' in line:
                            assert node['app_data']['ui_data']['label'] == read_key_pair(line)['value']
                        elif 'notebook=' in line:
                            assert component_parameters['filename'] == read_key_pair(line)['value']
                        elif 'image=' in line:
                            assert component_parameters['runtime_image'] == read_key_pair(line)['value']
                        elif 'env_vars=' in line:
                            for env in component_parameters['env_vars']:
                                var, value = env.split("=")
                                # Gets sub-list slice starting where the env vars starts
                                for env_line in file_as_lines[i + sub_list_line_counter + 2:]:
                                    if "AWS_ACCESS_KEY_ID" in env_line:
                                        assert sample_metadata['cos_username'] == read_key_pair(env_line,
                                                                                                sep=':')['value']
                                    elif "AWS_SECRET_ACCESS_KEY" in env_line:
                                        assert sample_metadata['cos_password'] == read_key_pair(env_line,
                                                                                                sep=':')['value']
                                    elif var in env_line:
                                        assert var == read_key_pair(env_line, sep=':')['key']
                                        assert value == read_key_pair(env_line, sep=':')['value']
                                    elif env_line.strip() == '},':  # end of env vars
                                        break
                        elif 'pipeline_inputs=' in line and component_parameters.get('inputs'):
                            for input in component_parameters['inputs']:
                                assert input in string_to_list(read_key_pair(line)['value'])
                        elif 'pipeline_outputs=' in line and component_parameters.get('outputs'):
                            for output in component_parameters['outputs']:
                                assert output in string_to_list(read_key_pair(line)['value'])
                        elif line == ')':  # End of this Notebook Op
                            break
                        sub_list_line_counter += 1


@pytest.mark.parametrize('parsed_pipeline', [PIPELINE_FILE_CUSTOM_COMPONENTS], indirect=True)
def test_create_file_custom_components(monkeypatch, processor, parsed_pipeline, parsed_ordered_dict, sample_metadata):
    pipeline_json = _read_pipeline_resource(PIPELINE_FILE_CUSTOM_COMPONENTS)

    export_pipeline_name = "some-name"
    export_file_type = "py"

    mocked_runtime = Metadata(name="test-metadata",
                              display_name="test",
                              schema_name="airflow",
                              metadata=sample_metadata
                              )

    monkeypatch.setattr(processor, "_get_metadata_configuration", lambda name=None, schemaspace=None: mocked_runtime)
    monkeypatch.setattr(processor, "_upload_dependencies_to_object_store", lambda x, y, z: True)
    monkeypatch.setattr(processor, "_cc_pipeline", lambda x, y: parsed_ordered_dict)

    with tempfile.TemporaryDirectory() as temp_dir:
        export_pipeline_output_path = os.path.join(temp_dir, f'{export_pipeline_name}.py')

        response = processor.create_pipeline_file(parsed_pipeline,
                                                  pipeline_export_format=export_file_type,
                                                  pipeline_export_path=export_pipeline_output_path,
                                                  pipeline_name=export_pipeline_name)

        assert export_pipeline_output_path == response
        assert os.path.isfile(export_pipeline_output_path)

        file_as_lines = open(response).read().splitlines()

        # Check DAG project name
        for i in range(len(file_as_lines)):
            if "args = {" == file_as_lines[i]:
                assert "project_id" == read_key_pair(file_as_lines[i + 1], sep=':')['key']
                assert export_pipeline_name == read_key_pair(file_as_lines[i + 1], sep=':')['value']

        # For every node in the original pipeline json
        for node in pipeline_json['pipelines'][0]['nodes']:
            component_parameters = node['app_data']['component_parameters']
            for i in range(len(file_as_lines)):
                # Matches custom component operators
                if f"op_{node['id'].replace('-', '_')} = " in file_as_lines[i]:
                    for parameter in component_parameters:
                        # Find 'parameter=' clause in file_as_lines list
                        r = re.compile(rf"\s*{parameter}=.*")
                        assert len(list(filter(r.match, file_as_lines[i + 1:]))) > 0


@pytest.mark.parametrize('parsed_pipeline', [PIPELINE_FILE_COMPLEX], indirect=True)
def test_export_overwrite(monkeypatch, processor, parsed_pipeline):
    with tempfile.TemporaryDirectory() as temp_dir:
        mocked_path = os.path.join(temp_dir, 'some-name.py')
        Path(mocked_path).touch()
        assert os.path.isfile(mocked_path)

        monkeypatch.setattr(processor, "create_pipeline_file",
                            lambda pipeline, pipeline_export_format, pipeline_export_path, pipeline_name: mocked_path)

        returned_path = processor.export(parsed_pipeline, "py", mocked_path, True)
        assert returned_path == mocked_path


@pytest.mark.parametrize('parsed_pipeline', [PIPELINE_FILE_COMPLEX], indirect=True)
def test_fail_export_overwrite(processor, parsed_pipeline):
    with tempfile.TemporaryDirectory() as temp_dir:
        Path(f"{temp_dir}/test.py").touch()
        assert os.path.isfile(f"{temp_dir}/test.py")

        export_pipeline_output_path = os.path.join(temp_dir, 'test.py')
        with pytest.raises(ValueError):
            processor.export(parsed_pipeline, "py", export_pipeline_output_path, False)


@pytest.mark.parametrize('parsed_pipeline', [PIPELINE_FILE_COMPLEX], indirect=True)
def test_pipeline_tree_creation(parsed_ordered_dict, sample_metadata, sample_image_metadata):
    pipeline_json = _read_pipeline_resource(PIPELINE_FILE_COMPLEX)

    ordered_dict = parsed_ordered_dict

    assert len(ordered_dict.keys()) == len(pipeline_json['pipelines'][0]['nodes'])

    # Verify tree structure is correct
    assert not ordered_dict['cded6818-e601-4fd8-b6b9-c9fdf1fd1fca'].get('parent_operation_ids')
    assert ordered_dict['bb9606ca-29ec-4133-a36a-67bd2a1f6dc3'].get(
        'parent_operation_ids').pop() == 'cded6818-e601-4fd8-b6b9-c9fdf1fd1fca'
    assert ordered_dict['6f5c2ece-1977-48a1-847f-099b327c6ed1'].get(
        'parent_operation_ids').pop() == 'cded6818-e601-4fd8-b6b9-c9fdf1fd1fca'
    assert ordered_dict['4ef63a48-a27c-4d1e-a0ee-2fbbdbe3be74'].get(
        'parent_operation_ids').pop() == 'cded6818-e601-4fd8-b6b9-c9fdf1fd1fca'
    assert ordered_dict['4f7ae91b-682e-476c-8664-58412336b31f'].get(
        'parent_operation_ids').pop() == 'bb9606ca-29ec-4133-a36a-67bd2a1f6dc3'
    assert ordered_dict['f82c4699-b392-4a3e-92b0-45d9e11126fe'].get(
        'parent_operation_ids').pop() == 'bb9606ca-29ec-4133-a36a-67bd2a1f6dc3'
    assert ordered_dict['137d3d2f-4224-42d9-b8c6-cbee9ff2872d'].get(
        'parent_operation_ids') == ['4ef63a48-a27c-4d1e-a0ee-2fbbdbe3be74', '0a7eff92-fe2a-411c-92a6-73d6f3810516']
    assert not ordered_dict['779c2630-64bf-47ca-8a98-9ac8a60e85f7'].get('parent_operation_ids')
    assert ordered_dict['0a7eff92-fe2a-411c-92a6-73d6f3810516'].get(
        'parent_operation_ids').pop() == '779c2630-64bf-47ca-8a98-9ac8a60e85f7'
    assert ordered_dict['92a7a247-1131-489c-8c3e-1e2389d4c673'].get(
        'parent_operation_ids') == ['f82c4699-b392-4a3e-92b0-45d9e11126fe',
                                    "137d3d2f-4224-42d9-b8c6-cbee9ff2872d",
                                    "6f5c2ece-1977-48a1-847f-099b327c6ed1"]

    for key in ordered_dict.keys():
        for node in pipeline_json['pipelines'][0]['nodes']:
            if node['id'] == key:
                component_parameters = node['app_data']['component_parameters']
                assert ordered_dict[key]['runtime_image'] == component_parameters['runtime_image']
                for image in sample_image_metadata:
                    if ordered_dict[key]['runtime_image'] == image.metadata['image_name']:
                        assert ordered_dict[key]['image_pull_policy'] == image.metadata['pull_policy']
                print(ordered_dict[key])
                for env in component_parameters['env_vars']:
                    var, value = env.split("=")
                    assert ordered_dict[key]['pipeline_envs'][var] == value
                assert ordered_dict[key]['pipeline_envs']['AWS_ACCESS_KEY_ID'] == sample_metadata['cos_username']
                assert ordered_dict[key]['pipeline_envs']['AWS_SECRET_ACCESS_KEY'] == sample_metadata['cos_password']
                for arg in ["inputs", "outputs"]:
                    if node['app_data'].get(arg):
                        for file in node['app_data'][arg]:
                            assert file in ordered_dict[key]["pipeline_" + arg]


def test_collect_envs(processor):
    pipelines_test_file = 'elyra/pipeline/tests/resources/archive/test.ipynb'

    # add system-owned envs with bogus values to ensure they get set to system-derived values,
    # and include some user-provided edge cases
    operation_envs = ['ELYRA_RUNTIME_ENV="bogus_runtime"',
                      'ELYRA_ENABLE_PIPELINE_INFO="bogus_pipeline"',
                      'ELYRA_WRITABLE_CONTAINER_DIR=',  # simulate operation reference in pipeline
                      'AWS_ACCESS_KEY_ID="bogus_key"',
                      'AWS_SECRET_ACCESS_KEY="bogus_secret"',
                      'USER_EMPTY_VALUE=  ',
                      'USER_TWO_EQUALS=KEY=value',
                      'USER_NO_VALUE=']
    component_parameters = {
        'filename': pipelines_test_file,
        'env_vars': operation_envs,
        'runtime_image': 'tensorflow/tensorflow:latest'
    }
    test_operation = GenericOperation(id='this-is-a-test-id',
                                      type='execution-node',
                                      classifier='execute-notebook-node',
                                      name='test',
                                      component_params=component_parameters)

    envs = processor._collect_envs(test_operation, cos_secret=None, cos_username='Alice', cos_password='secret')

    assert envs['ELYRA_RUNTIME_ENV'] == 'airflow'
    assert envs['AWS_ACCESS_KEY_ID'] == 'Alice'
    assert envs['AWS_SECRET_ACCESS_KEY'] == 'secret'
    assert envs['ELYRA_ENABLE_PIPELINE_INFO'] == 'True'
    assert 'ELYRA_WRITABLE_CONTAINER_DIR' not in envs
    assert envs['USER_EMPTY_VALUE'] == '  '
    assert envs['USER_TWO_EQUALS'] == 'KEY=value'
    assert 'USER_NO_VALUE' not in envs

    # Repeat with non-None secret - ensure user and password envs are not present, but others are
    envs = processor._collect_envs(test_operation, cos_secret='secret', cos_username='Alice', cos_password='secret')

    assert envs['ELYRA_RUNTIME_ENV'] == 'airflow'
    assert 'AWS_ACCESS_KEY_ID' not in envs
    assert 'AWS_SECRET_ACCESS_KEY' not in envs
    assert envs['ELYRA_ENABLE_PIPELINE_INFO'] == 'True'
    assert 'ELYRA_WRITABLE_CONTAINER_DIR' not in envs
    assert envs['USER_EMPTY_VALUE'] == '  '
    assert envs['USER_TWO_EQUALS'] == 'KEY=value'
    assert 'USER_NO_VALUE' not in envs


def test_unique_operation_name_existent(processor):
    operation_name = "sample_operation"

    op1 = {'notebook': "sample_operation"}
    op2 = {'notebook': "sample_operation_2"}
    op3 = {'notebook': "sample_operation_3"}
    sample_operation_list = [op1, op2, op3]

    unique_name = processor._get_unique_operation_name(operation_name, sample_operation_list)

    assert unique_name == "sample_operation_4"


def test_unique_operation_name_non_existent(processor):
    operation_name = "sample_operation_foo_bar"

    op1 = {'notebook': "sample_operation"}
    op2 = {'notebook': "sample_operation_2"}
    op3 = {'notebook': "sample_operation_3"}
    sample_operation_list = [op1, op2, op3]

    unique_name = processor._get_unique_operation_name(operation_name, sample_operation_list)

    assert unique_name == operation_name


def test_process_list_value_function(processor):
    # Test values that will be successfully converted to list
    assert processor._process_list_value("") == []
    assert processor._process_list_value(None) == []
    assert processor._process_list_value("[]") == []
    assert processor._process_list_value("None") == []
    assert processor._process_list_value("['elem1']") == ["elem1"]
    assert processor._process_list_value("['elem1', 'elem2', 'elem3']") == ["elem1", "elem2", "elem3"]
    assert processor._process_list_value("  ['elem1',   'elem2' , 'elem3']  ") == ["elem1", "elem2", "elem3"]
    assert processor._process_list_value("[1, 2]") == [1, 2]
    assert processor._process_list_value("[True, False, True]") == [True, False, True]
    assert processor._process_list_value("[{'obj': 'val', 'obj2': 'val2'}, {}]") == [{'obj': 'val', 'obj2': 'val2'}, {}]

    # Test values that will not be successfully converted to list
    # Surrounding quotes are added to string values for correct DAG render
    assert processor._process_list_value("[[]") == "\"[[]\""
    assert processor._process_list_value("[elem1, elem2]") == "\"[elem1, elem2]\""
    assert processor._process_list_value("elem1, elem2") == "\"elem1, elem2\""
    assert processor._process_list_value("  elem1, elem2  ") == "\"elem1, elem2\""
    assert processor._process_list_value("'elem1', 'elem2'") == "\"'elem1', 'elem2'\""


def test_process_dictionary_value_function(processor):
    # Test values that will be successfully converted to dictionary
    assert processor._process_dictionary_value("") == {}
    assert processor._process_dictionary_value(None) == {}
    assert processor._process_dictionary_value("{}") == {}
    assert processor._process_dictionary_value("None") == {}
    assert processor._process_dictionary_value("{'key': 'value'}") == {"key": "value"}

    dict_as_str = "{'key1': 'value', 'key2': 'value'}"
    assert processor._process_dictionary_value(dict_as_str) == {"key1": "value", "key2": "value"}

    dict_as_str = "  {  'key1': 'value'  , 'key2'  : 'value'}  "
    assert processor._process_dictionary_value(dict_as_str) == {"key1": "value", "key2": "value"}

    dict_as_str = "{'key1': [1, 2, 3], 'key2': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == {"key1": [1, 2, 3], "key2": ["elem1", "elem2"]}

    dict_as_str = "{'key1': 2, 'key2': 'value', 'key3': True, 'key4': None, 'key5': [1, 2, 3]}"
    expected_value = {
        "key1": 2,
        "key2": "value",
        "key3": True,
        "key4": None,
        "key5": [1, 2, 3]
    }
    assert processor._process_dictionary_value(dict_as_str) == expected_value

    dict_as_str = "{'key1': {'key2': 2, 'key3': 3, 'key4': 4}, 'key5': {}}"
    expected_value = {
        "key1": {
            "key2": 2,
            "key3": 3,
            "key4": 4,
        },
        "key5": {}
    }
    assert processor._process_dictionary_value(dict_as_str) == expected_value

    # Test values that will not be successfully converted to dictionary
    # Surrounding quotes are added to string values for correct DAG render
    assert processor._process_dictionary_value("{{}") == "\"{{}\""
    assert processor._process_dictionary_value("{key1: value, key2: value}") == "\"{key1: value, key2: value}\""
    assert processor._process_dictionary_value("  { key1: value, key2: value }  ") == "\"{ key1: value, key2: value }\""
    assert processor._process_dictionary_value("key1: value, key2: value") == "\"key1: value, key2: value\""
    assert processor._process_dictionary_value("{'key1': true}") == "\"{'key1': true}\""
    assert processor._process_dictionary_value("{'key': null}") == "\"{'key': null}\""

    dict_as_str = "{'key1': [elem1, elem2, elem3], 'key2': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == f"\"{dict_as_str}\""

    dict_as_str = "{'key1': {key2: 2}, 'key3': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == f"\"{dict_as_str}\""

    dict_as_str = "{'key1': {key2: 2}, 'key3': ['elem1', 'elem2']}"
    assert processor._process_dictionary_value(dict_as_str) == f"\"{dict_as_str}\""



#
# Copyright 2018-2021 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

from elyra.metadata.manager import MetadataManager
from elyra.metadata.schema import METADATA_TEST_SCHEMASPACE
from elyra.metadata.schema import SchemaManager
from elyra.tests.metadata.test_utils import another_metadata_json
from elyra.tests.metadata.test_utils import byo_metadata_json
from elyra.tests.metadata.test_utils import create_instance
from elyra.tests.metadata.test_utils import create_json_file
from elyra.tests.metadata.test_utils import invalid_json
from elyra.tests.metadata.test_utils import invalid_metadata_json
from elyra.tests.metadata.test_utils import invalid_schema_name_json
from elyra.tests.metadata.test_utils import valid_metadata_json


def mkdir(tmp_path, *parts):
    path = tmp_path.joinpath(*parts)
    if not path.exists():
        path.mkdir(parents=True)
    return path


# These location fixtures will need to be revisited once we support multiple metadata storage types.
schemaspace_location = pytest.fixture(lambda jp_data_dir:
                                      mkdir(jp_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
bogus_location = pytest.fixture(lambda jp_data_dir:
                                mkdir(jp_data_dir, "metadata", "bogus"))
shared_location = pytest.fixture(lambda jp_system_jupyter_path:
                                 mkdir(jp_system_jupyter_path, "metadata", METADATA_TEST_SCHEMASPACE))
factory_location = pytest.fixture(lambda jp_env_jupyter_path:
                                  mkdir(jp_env_jupyter_path, "metadata", METADATA_TEST_SCHEMASPACE))


@pytest.fixture
def setup_data(schemaspace_location):
    create_json_file(schemaspace_location, 'valid.json', valid_metadata_json)
    create_json_file(schemaspace_location, 'another.json', another_metadata_json)
    create_json_file(schemaspace_location, 'invalid.json', invalid_metadata_json)


@pytest.fixture
def setup_hierarchy(jp_environ, factory_location):
    # Only populate factory info
    byo_instance = byo_metadata_json
    byo_instance['display_name'] = 'factory'
    create_json_file(factory_location, 'byo_1.json', byo_instance)
    create_json_file(factory_location, 'byo_2.json', byo_instance)
    create_json_file(factory_location, 'byo_3.json', byo_instance)


@pytest.fixture
def store_manager(tests_manager):
    return tests_manager.metadata_store


@pytest.fixture(params=["elyra.metadata.storage.FileMetadataStore",
                        "elyra.tests.metadata.test_utils.MockMetadataStore"])  # Add types as needed
def tests_manager(jp_environ, schemaspace_location, request):
    metadata_mgr = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE, metadata_store_class=request.param)
    store_mgr = metadata_mgr.metadata_store
    create_instance(store_mgr, schemaspace_location, 'valid', valid_metadata_json)
    create_instance(store_mgr, schemaspace_location, 'another', another_metadata_json)
    create_instance(store_mgr, schemaspace_location, 'invalid', invalid_metadata_json)
    create_instance(store_mgr, schemaspace_location, 'bad', invalid_json)
    create_instance(store_mgr, schemaspace_location, 'invalid_schema_name', invalid_schema_name_json)
    return metadata_mgr


@pytest.fixture
def tests_hierarchy_manager(setup_hierarchy):  # Only uses FileMetadataStore for storage right now.
    return MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)


@pytest.fixture
def schema_manager():
    schema_manager = SchemaManager.instance()
    yield schema_manager
    SchemaManager.clear_instance()


# Set Elyra server extension as enabled (overriding server_config fixture from jupyter_server)
@pytest.fixture
def jp_server_config():
    return {
        "ServerApp": {
            "jpserver_extensions": {
                "elyra": True
            }
        }
    }
